# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the migrations."""

import hashlib
from typing import Any, cast, override

from django.conf import settings
from django.db import connection
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.state import StateApps
from django.test import TestCase
from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    TaskTypes,
)
from debusine.client.models import RuntimeParameter, RuntimeParameters
from debusine.db.models import (
    DEFAULT_FILE_STORE_NAME,
    FileStore,
    User,
    WorkRequest,
)
from debusine.db.models.auth import _TokenTypes
from debusine.tasks.models import WorkerType


def _migrate_check_constraints(action: str, *args: Any, **kwargs: Any) -> None:
    """
    Check constraints before and after (un)applying each migration.

    When applying migrations in tests, each migration only ends with a
    savepoint rather than a full commit, so deferred constraints aren't
    immediately checked.  As a result, we need to explicitly check
    constraints to avoid "cannot ALTER TABLE ... because it has pending
    trigger events" errors in some cases.
    """
    if action in {
        "apply_start",
        "apply_success",
        "unapply_start",
        "unapply_success",
    }:
        connection.check_constraints()


LATEST_IRREVERSIBLE_MIGRATION = [("db", "0001_squashed_0_11_0")]


class MigrationTests(TestCase):
    """Test migrations."""

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        # It seems to be faster to apply most of the migrations forwards
        # rather than backwards.  Migrate back as far as we can before
        # running this test suite.
        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(LATEST_IRREVERSIBLE_MIGRATION)

    @override
    @classmethod
    def tearDownClass(cls) -> None:
        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(executor.loader.graph.leaf_nodes())
        super().tearDownClass()

    @staticmethod
    def get_test_user_for_apps(apps: StateApps) -> User:
        """
        Return a test user.

        We may be testing a migration, so use the provided application state.
        """
        user_model = apps.get_model("db", "User")
        try:
            user = user_model.objects.get(username="usertest")
        except user_model.DoesNotExist:
            user = user_model.objects.create_user(
                username="usertest",
                password="userpassword",
                email="usertest@example.org",
            )
        return cast(User, user)

    def test_default_store_created(self) -> None:
        """Assert Default FileStore has been created."""
        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(LATEST_IRREVERSIBLE_MIGRATION)
        apps = executor.loader.project_state(LATEST_IRREVERSIBLE_MIGRATION).apps

        default_file_store = apps.get_model("db", "FileStore").objects.get(
            name=DEFAULT_FILE_STORE_NAME
        )

        self.assertEqual(
            default_file_store.backend, FileStore.BackendChoices.LOCAL
        )

        self.assertEqual(default_file_store.configuration, {})

    def test_system_workspace_created(self) -> None:
        """Assert System Workspace has been created."""
        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(LATEST_IRREVERSIBLE_MIGRATION)
        apps = executor.loader.project_state(LATEST_IRREVERSIBLE_MIGRATION).apps

        workspace = apps.get_model("db", "Workspace").objects.get(
            scope__name=settings.DEBUSINE_DEFAULT_SCOPE,
            name=settings.DEBUSINE_DEFAULT_WORKSPACE,
        )
        default_file_store = apps.get_model("db", "FileStore").objects.get(
            name=DEFAULT_FILE_STORE_NAME
        )
        self.assertQuerySetEqual(
            workspace.scope.file_stores.all(), [default_file_store]
        )

    def assert_work_request_task_data_renamed(
        self,
        migrate_from: str,
        migrate_to: str,
        old_work_requests: list[tuple[TaskTypes, str, dict[str, Any]]],
        new_work_requests: list[tuple[TaskTypes, str, dict[str, Any]]],
    ) -> None:
        """Assert that migrations rename task data in work requests."""
        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate([("db", migrate_from)])
        old_apps = executor.loader.project_state([("db", migrate_from)]).apps
        workspace = old_apps.get_model("db", "Workspace").objects.get(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        for task_type, task_name, task_data in old_work_requests:
            old_apps.get_model("db", "WorkRequest").objects.create(
                created_by_id=self.get_test_user_for_apps(old_apps).id,
                task_type=task_type,
                task_name=task_name,
                workspace=workspace,
                task_data=task_data,
            )

        executor.loader.build_graph()
        executor.migrate([("db", migrate_to)])
        new_apps = executor.loader.project_state([("db", migrate_to)]).apps

        self.assertCountEqual(
            new_apps.get_model("db", "WorkRequest").objects.values_list(
                "task_type", "task_name", "task_data"
            ),
            new_work_requests,
        )

        executor.loader.build_graph()
        executor.migrate([("db", migrate_from)])

        self.assertCountEqual(
            old_apps.get_model("db", "WorkRequest").objects.values_list(
                "task_type", "task_name", "task_data"
            ),
            old_work_requests,
        )

    def assert_work_request_all_task_data_renamed(
        self,
        migrate_from: str,
        migrate_to: str,
        old_work_requests: list[
            tuple[
                TaskTypes,
                str,
                dict[str, Any],
                dict[str, Any],
                dict[str, Any] | None,
            ]
        ],
        new_work_requests: list[
            tuple[
                TaskTypes,
                str,
                dict[str, Any],
                dict[str, Any],
                dict[str, Any] | None,
            ]
        ],
    ) -> None:
        """Assert that migrations rename all task data in work requests."""
        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate([("db", migrate_from)])
        old_apps = executor.loader.project_state([("db", migrate_from)]).apps
        workspace = old_apps.get_model("db", "Workspace").objects.get(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        for (
            task_type,
            task_name,
            task_data,
            dynamic_task_data,
            configured_task_data,
        ) in old_work_requests:
            old_apps.get_model("db", "WorkRequest").objects.create(
                created_by_id=self.get_test_user_for_apps(old_apps).id,
                task_type=task_type,
                task_name=task_name,
                workspace=workspace,
                task_data=task_data,
                dynamic_task_data=dynamic_task_data,
                configured_task_data=configured_task_data,
            )

        executor.loader.build_graph()
        executor.migrate([("db", migrate_to)])
        new_apps = executor.loader.project_state([("db", migrate_to)]).apps

        self.assertCountEqual(
            new_apps.get_model("db", "WorkRequest").objects.values_list(
                "task_type",
                "task_name",
                "task_data",
                "dynamic_task_data",
                "configured_task_data",
            ),
            new_work_requests,
        )

        executor.loader.build_graph()
        executor.migrate([("db", migrate_from)])

        self.assertCountEqual(
            old_apps.get_model("db", "WorkRequest").objects.values_list(
                "task_type",
                "task_name",
                "task_data",
                "dynamic_task_data",
                "configured_task_data",
            ),
            old_work_requests,
        )

    def assert_workflow_template_task_data_renamed(
        self,
        migrate_from: str,
        migrate_to: str,
        old_workflow_templates: list[tuple[str, dict[str, Any]]],
        new_workflow_templates: list[tuple[str, dict[str, Any]]],
    ) -> None:
        """Assert that migrations rename task data in workflow templates."""
        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate([("db", migrate_from)])
        old_apps = executor.loader.project_state([("db", migrate_from)]).apps
        workspace = old_apps.get_model("db", "Workspace").objects.get(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        for i, (task_name, task_data) in enumerate(old_workflow_templates):
            old_apps.get_model("db", "WorkflowTemplate").objects.create(
                name=f"rename-test-{i}",
                workspace=workspace,
                task_name=task_name,
                task_data=task_data,
            )

        executor.loader.build_graph()
        executor.migrate([("db", migrate_to)])
        new_apps = executor.loader.project_state([("db", migrate_to)]).apps

        self.assertCountEqual(
            new_apps.get_model("db", "WorkflowTemplate").objects.values_list(
                "name", "task_name", "task_data"
            ),
            [
                (f"rename-test-{i}", task_name, task_data)
                for i, (task_name, task_data) in enumerate(
                    new_workflow_templates
                )
            ],
        )

        executor.loader.build_graph()
        executor.migrate([("db", migrate_from)])

        self.assertCountEqual(
            old_apps.get_model("db", "WorkflowTemplate").objects.values_list(
                "name", "task_name", "task_data"
            ),
            [
                (f"rename-test-{i}", task_name, task_data)
                for i, (task_name, task_data) in enumerate(
                    old_workflow_templates
                )
            ],
        )

    def assert_workflow_template_parameters_renamed(
        self,
        migrate_from: str,
        migrate_to: str,
        old_workflow_templates: list[
            tuple[str, dict[str, Any], RuntimeParameters]
        ],
        new_workflow_templates: list[
            tuple[str, dict[str, Any], RuntimeParameters]
        ],
    ) -> None:
        """Assert that migrations rename parameters in workflow templates."""
        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate([("db", migrate_from)])
        old_apps = executor.loader.project_state([("db", migrate_from)]).apps
        workspace = old_apps.get_model("db", "Workspace").objects.get(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        for i, (task_name, static_parameters, runtime_parameters) in enumerate(
            old_workflow_templates
        ):
            old_apps.get_model("db", "WorkflowTemplate").objects.create(
                name=f"rename-test-{i}",
                workspace=workspace,
                task_name=task_name,
                static_parameters=static_parameters,
                runtime_parameters=runtime_parameters,
            )

        executor.loader.build_graph()
        executor.migrate([("db", migrate_to)])
        new_apps = executor.loader.project_state([("db", migrate_to)]).apps

        self.assertCountEqual(
            new_apps.get_model("db", "WorkflowTemplate").objects.values_list(
                "name", "task_name", "static_parameters", "runtime_parameters"
            ),
            [
                (
                    f"rename-test-{i}",
                    task_name,
                    static_parameters,
                    runtime_parameters,
                )
                for i, (
                    task_name,
                    static_parameters,
                    runtime_parameters,
                ) in enumerate(new_workflow_templates)
            ],
        )

        executor.loader.build_graph()
        executor.migrate([("db", migrate_from)])

        self.assertCountEqual(
            old_apps.get_model("db", "WorkflowTemplate").objects.values_list(
                "name", "task_name", "static_parameters", "runtime_parameters"
            ),
            [
                (
                    f"rename-test-{i}",
                    task_name,
                    static_parameters,
                    runtime_parameters,
                )
                for i, (
                    task_name,
                    static_parameters,
                    runtime_parameters,
                ) in enumerate(old_workflow_templates)
            ],
        )

    def test_artifact_system_components(self) -> None:
        """Test adding components to debian:system-tarball artifacts."""
        migrate_from = [("db", "0003_add_workspace_role_viewer")]
        migrate_to = [("db", "0004_system_image_components")]

        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(migrate_from)
        old_apps = executor.loader.project_state(migrate_from).apps
        workspace = old_apps.get_model("db", "Workspace").objects.get(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        OldArtifact = old_apps.get_model("db", "Artifact")
        OldWorkRequest = old_apps.get_model("db", "WorkRequest")
        wr_unrelated = OldWorkRequest.objects.create(
            created_by_id=self.get_test_user_for_apps(old_apps).id,
            task_name="noop",
            workspace=workspace,
            task_data={},
        )
        wr_without_components = OldWorkRequest.objects.create(
            created_by_id=self.get_test_user_for_apps(old_apps).id,
            task_name="mmdebstrap",
            workspace=workspace,
            task_data={
                "bootstrap_repositories": [
                    {
                        "mirror": "http://deb.debian.org/debian",
                        "suite": "unstable",
                    },
                ],
            },
        )
        wr_with_components = OldWorkRequest.objects.create(
            created_by_id=self.get_test_user_for_apps(old_apps).id,
            task_name="mmdebstrap",
            workspace=workspace,
            task_data={
                "bootstrap_repositories": [
                    {
                        "mirror": "http://deb.debian.org/debian",
                        "suite": "unstable",
                        "components": ["main", "contrib"],
                    },
                ],
            },
        )

        for data, created_by_work_request in (
            ({"components": ["foo"]}, None),
            ({}, None),
            ({}, wr_unrelated),
            ({}, wr_without_components),
            ({}, wr_with_components),
        ):
            OldArtifact.objects.create(
                category=ArtifactCategory.SYSTEM_TARBALL,
                data=data,
                created_by_work_request=created_by_work_request,
                workspace=workspace,
            )

        executor.loader.build_graph()
        executor.migrate(migrate_to)

        new_apps = executor.loader.project_state(migrate_to).apps
        NewArtifact = new_apps.get_model("db", "Artifact")

        self.assertQuerySetEqual(
            NewArtifact.objects.order_by("id").values_list(
                "data__components", flat=True
            ),
            [
                ["foo"],  # already specified
                [],
                [],
                [],
                ["main", "contrib"],  # looked up from components
            ],
        )

        executor.loader.build_graph()
        executor.migrate(migrate_from)

        self.assertFalse(
            OldArtifact.objects.filter(data__components__isnull=False).exists()
        )

    def test_create_archive(self) -> None:
        """``debian:archive`` is created in the default workspace."""
        migrate_from = [("db", "0009_add_archive_singleton_collection")]
        migrate_to = [("db", "0010_add_archive_singleton_collection_data")]

        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(migrate_from)
        old_apps = executor.loader.project_state(migrate_from).apps
        OldWorkspace = old_apps.get_model("db", "Workspace")
        OldCollection = old_apps.get_model("db", "Collection")
        self.assertQuerySetEqual(
            OldCollection.objects.filter(
                name="_",
                category=CollectionCategory.ARCHIVE,
                workspace__scope__name=settings.DEBUSINE_DEFAULT_SCOPE,
                workspace__name=settings.DEBUSINE_DEFAULT_WORKSPACE,
            ),
            [],
        )

        default_workspace = OldWorkspace.objects.get(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        other_workspace = OldWorkspace.objects.create(
            scope=default_workspace.scope, name="other"
        )
        OldCollection.objects.create(
            name="bookworm",
            category=CollectionCategory.SUITE,
            workspace=default_workspace,
        )
        OldCollection.objects.create(
            name="other",
            category=CollectionCategory.SUITE,
            workspace=other_workspace,
        )

        executor.loader.build_graph()
        executor.migrate(migrate_to)
        new_apps = executor.loader.project_state(migrate_to).apps
        new_archive = new_apps.get_model("db", "Collection").objects.get(
            name="_",
            category=CollectionCategory.ARCHIVE,
            workspace__scope__name=settings.DEBUSINE_DEFAULT_SCOPE,
            workspace__name=settings.DEBUSINE_DEFAULT_WORKSPACE,
        )
        self.assertQuerySetEqual(
            new_archive.child_items.values_list(
                "name", "parent_category", "category", "collection__name"
            ),
            [
                (
                    "bookworm",
                    CollectionCategory.ARCHIVE,
                    CollectionCategory.SUITE,
                    "bookworm",
                )
            ],
        )

        executor.loader.build_graph()
        executor.migrate(migrate_from)

        self.assertQuerySetEqual(
            OldCollection.objects.filter(
                name="_",
                category=CollectionCategory.ARCHIVE,
                workspace__scope__name=settings.DEBUSINE_DEFAULT_SCOPE,
                workspace__name=settings.DEBUSINE_DEFAULT_WORKSPACE,
            ),
            [],
        )

    def test_qa_suite_rename(self) -> None:
        """Various workflow data fields are renamed to ``qa_suite``."""
        self.assert_work_request_task_data_renamed(
            migrate_from=(
                "0012_alter_collectionitem_created_by_workflow_and_more"
            ),
            migrate_to="0013_qa_suite",
            old_work_requests=[
                (
                    TaskTypes.WORKFLOW,
                    "foo",
                    {
                        "suite_collection": "bookworm@debian:suite",
                        "reverse_dependencies_autopkgtest_suite": (
                            "bookworm@debian:suite"
                        ),
                    },
                ),
                (
                    TaskTypes.WORKFLOW,
                    "reverse_dependencies_autopkgtest",
                    {"suite_collection": "bookworm@debian:suite"},
                ),
                (
                    TaskTypes.WORKFLOW,
                    "qa",
                    {
                        "reverse_dependencies_autopkgtest_suite": (
                            "bookworm@debian:suite"
                        )
                    },
                ),
                (
                    TaskTypes.WORKFLOW,
                    "debian_pipeline",
                    {
                        "reverse_dependencies_autopkgtest_suite": (
                            "bookworm@debian:suite"
                        )
                    },
                ),
            ],
            new_work_requests=[
                (
                    TaskTypes.WORKFLOW,
                    "foo",
                    {
                        "suite_collection": "bookworm@debian:suite",
                        "reverse_dependencies_autopkgtest_suite": (
                            "bookworm@debian:suite"
                        ),
                    },
                ),
                (
                    TaskTypes.WORKFLOW,
                    "reverse_dependencies_autopkgtest",
                    {"qa_suite": "bookworm@debian:suite"},
                ),
                (
                    TaskTypes.WORKFLOW,
                    "qa",
                    {"qa_suite": "bookworm@debian:suite"},
                ),
                (
                    TaskTypes.WORKFLOW,
                    "debian_pipeline",
                    {"qa_suite": "bookworm@debian:suite"},
                ),
            ],
        )
        self.assert_workflow_template_task_data_renamed(
            migrate_from=(
                "0012_alter_collectionitem_created_by_workflow_and_more"
            ),
            migrate_to="0013_qa_suite",
            old_workflow_templates=[
                (
                    "foo",
                    {
                        "suite_collection": "bookworm@debian:suite",
                        "reverse_dependencies_autopkgtest_suite": (
                            "bookworm@debian:suite"
                        ),
                    },
                ),
                (
                    "reverse_dependencies_autopkgtest",
                    {"suite_collection": "bookworm@debian:suite"},
                ),
                (
                    "qa",
                    {
                        "reverse_dependencies_autopkgtest_suite": (
                            "bookworm@debian:suite"
                        )
                    },
                ),
                (
                    "debian_pipeline",
                    {
                        "reverse_dependencies_autopkgtest_suite": (
                            "bookworm@debian:suite"
                        )
                    },
                ),
            ],
            new_workflow_templates=[
                (
                    "foo",
                    {
                        "suite_collection": "bookworm@debian:suite",
                        "reverse_dependencies_autopkgtest_suite": (
                            "bookworm@debian:suite"
                        ),
                    },
                ),
                (
                    "reverse_dependencies_autopkgtest",
                    {"qa_suite": "bookworm@debian:suite"},
                ),
                ("qa", {"qa_suite": "bookworm@debian:suite"}),
                ("debian_pipeline", {"qa_suite": "bookworm@debian:suite"}),
            ],
        )

    def test_moving_task_configuration_id(self) -> None:
        """Test moving task_configuration_id from dynamic to configured data."""
        migrate_from = [
            ("db", "0019_workflowtemplate_db_workflowtemplate_update_suites")
        ]
        migrate_to = [("db", "0020_move_task_configuration_id")]

        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(migrate_from)

        old_apps = executor.loader.project_state(migrate_from).apps
        workspace = old_apps.get_model("db", "Workspace").objects.get(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        OldWorkRequest = old_apps.get_model("db", "WorkRequest")

        common_kwargs = {
            "created_by_id": self.get_test_user_for_apps(old_apps).id,
            "task_name": "noop",
            "workspace": workspace,
            "task_data": {},
        }

        pre: dict[str, WorkRequest] = {}

        pre["no_id1"] = OldWorkRequest.objects.create(**common_kwargs)
        pre["no_id2"] = OldWorkRequest.objects.create(
            dynamic_task_data={}, **common_kwargs
        )
        pre["no_id3"] = OldWorkRequest.objects.create(
            configured_task_data={}, **common_kwargs
        )
        pre["no_id4"] = OldWorkRequest.objects.create(
            dynamic_task_data={}, configured_task_data={}, **common_kwargs
        )
        pre["no_id5"] = OldWorkRequest.objects.create(
            dynamic_task_data={"test": 1},
            configured_task_data={"test": 2},
            **common_kwargs,
        )

        pre["null_id1"] = OldWorkRequest.objects.create(
            dynamic_task_data={"task_configuration_id": None}, **common_kwargs
        )
        pre["null_id2"] = OldWorkRequest.objects.create(
            dynamic_task_data={"task_configuration_id": None},
            configured_task_data={},
            **common_kwargs,
        )
        pre["null_id3"] = OldWorkRequest.objects.create(
            dynamic_task_data={"task_configuration_id": None},
            configured_task_data={"task_configuration": None},
            **common_kwargs,
        )
        pre["null_id4"] = OldWorkRequest.objects.create(
            dynamic_task_data={"task_configuration_id": None},
            configured_task_data={"task_configuration": 3},
            **common_kwargs,
        )
        pre["null_id5"] = OldWorkRequest.objects.create(
            dynamic_task_data={"task_configuration_id": None, "test": 1},
            configured_task_data={"task_configuration": 3, "test": 2},
            **common_kwargs,
        )

        pre["valid_id1"] = OldWorkRequest.objects.create(
            dynamic_task_data={"task_configuration_id": 3}, **common_kwargs
        )
        pre["valid_id2"] = OldWorkRequest.objects.create(
            dynamic_task_data={"task_configuration_id": 3},
            configured_task_data={},
            **common_kwargs,
        )
        pre["valid_id3"] = OldWorkRequest.objects.create(
            dynamic_task_data={"task_configuration_id": 3},
            configured_task_data={"task_configuration": None},
            **common_kwargs,
        )
        pre["valid_id4"] = OldWorkRequest.objects.create(
            dynamic_task_data={"task_configuration_id": 3},
            configured_task_data={"task_configuration": 1},
            **common_kwargs,
        )
        pre["valid_id5"] = OldWorkRequest.objects.create(
            dynamic_task_data={"task_configuration_id": 3, "test": 1},
            configured_task_data={"task_configuration": 1, "test": 2},
            **common_kwargs,
        )

        executor.loader.build_graph()
        executor.migrate(migrate_to)

        new_apps = executor.loader.project_state(migrate_to).apps
        NewWorkRequest = new_apps.get_model("db", "WorkRequest")

        for name, expected_configured in (
            ("no_id1", None),
            ("no_id2", None),
            ("no_id3", {}),
            ("no_id4", {}),
            ("no_id5", {"test": 2}),
            ("null_id1", None),
            ("null_id2", {}),
            ("null_id3", {"task_configuration": None}),
            ("null_id4", {"task_configuration": 3}),
            ("null_id5", {"task_configuration": 3, "test": 2}),
            ("valid_id1", {"task_configuration": 3}),
            ("valid_id2", {"task_configuration": 3}),
            ("valid_id3", {"task_configuration": 3}),
            ("valid_id4", {"task_configuration": 1}),
            ("valid_id5", {"task_configuration": 1, "test": 2}),
        ):
            with self.subTest(name=name):
                old = pre[name]
                new = NewWorkRequest.objects.get(pk=old.pk)
                if new.dynamic_task_data is not None:
                    self.assertNotIn(
                        "task_configuration_id", new.dynamic_task_data
                    )
                    old_dynamic = (old.dynamic_task_data or {}).copy()
                    old_dynamic.pop("task_configuration_id", None)
                    self.assertEqual(new.dynamic_task_data, old_dynamic)
                self.assertEqual(new.configured_task_data, expected_configured)

        executor.loader.build_graph()
        executor.migrate(migrate_from)

        for name, expected_dynamic in (
            ("no_id1", None),
            ("no_id2", {}),
            ("no_id3", None),
            ("no_id4", {}),
            ("no_id5", {"test": 1}),
            ("null_id1", {}),
            ("null_id2", {}),
            ("null_id3", {}),
            ("null_id4", {"task_configuration_id": 3}),
            ("null_id5", {"task_configuration_id": 3, "test": 1}),
            ("valid_id1", {"task_configuration_id": 3}),
            ("valid_id2", {"task_configuration_id": 3}),
            ("valid_id3", {"task_configuration_id": 3}),
            ("valid_id4", {"task_configuration_id": 1}),
            ("valid_id5", {"task_configuration_id": 1, "test": 1}),
        ):
            with self.subTest(name=name):
                old = pre[name]
                newold = OldWorkRequest.objects.get(pk=old.pk)
                if newold.configured_task_data is not None:
                    self.assertNotIn(
                        "task_configuration", newold.configured_task_data
                    )
                    old_configured = (old.configured_task_data or {}).copy()
                    old_configured.pop("task_configuration", None)
                    self.assertEqual(
                        newold.configured_task_data, old_configured
                    )
                self.assertEqual(newold.dynamic_task_data, expected_dynamic)

    def test_work_request_set_workflow_root(self) -> None:
        migrate_from = [("db", "0022_workrequest_workflow_root")]
        migrate_to = [("db", "0023_workrequest_workflow_root_data")]

        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(migrate_from)

        old_apps = executor.loader.project_state(migrate_from).apps
        workspace = old_apps.get_model("db", "Workspace").objects.get(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        OldWorkRequest = old_apps.get_model("db", "WorkRequest")

        common_kwargs = {
            "created_by_id": self.get_test_user_for_apps(old_apps).id,
            "task_name": "noop",
            "workspace": workspace,
            "task_data": {},
        }

        pre: dict[str, WorkRequest] = {}
        expected_root_pks: dict[str, int | None] = {}

        for path, is_workflow, expected_root_path in (
            ("1", True, "1"),
            ("1-1", True, "1"),
            ("1-1-1", False, "1"),
            ("1-1-2", True, "1"),
            ("1-1-2-1", False, "1"),
            ("1-1-3", False, "1"),
            ("1-2", True, "1"),
            ("1-2-1", True, "1"),
            ("1-2-1-1", False, "1"),
            ("2", True, "2"),
            ("2-1", True, "2"),
            ("2-1-1", True, "2"),
            ("2-1-1-1", False, "2"),
            ("2-2", True, "2"),
            ("3", True, "3"),
            ("3-1", False, "3"),
            ("4", True, "4"),
            ("5", False, None),
        ):
            parent = pre.get("-".join(path.split("-")[:-1]))
            if parent is not None:
                self.assertEqual(parent.task_type, TaskTypes.WORKFLOW)
            pre[path] = OldWorkRequest.objects.create(
                task_type=(
                    TaskTypes.WORKFLOW if is_workflow else TaskTypes.WORKER
                ),
                parent=pre.get("-".join(path.split("-")[:-1])),
                **common_kwargs,
            )
            expected_root_pks[path] = (
                None
                if expected_root_path is None
                else pre[expected_root_path].pk
            )

        executor.loader.build_graph()
        executor.migrate(migrate_to)

        new_apps = executor.loader.project_state(migrate_to).apps
        NewWorkRequest = new_apps.get_model("db", "WorkRequest")

        for path, expected_root_pk in expected_root_pks.items():
            with self.subTest(path=path, state="after"):
                self.assertEqual(
                    NewWorkRequest.objects.get(
                        pk=pre[path].pk
                    ).workflow_root_id,
                    expected_root_pk,
                )

        executor.loader.build_graph()
        executor.migrate(migrate_from)

        for path in pre:
            with self.subTest(path=path, state="before"):
                self.assertIsNone(
                    OldWorkRequest.objects.get(pk=pre[path].pk).workflow_root_id
                )

    def test_rename_publish_target_to_suite(self) -> None:
        self.assert_work_request_all_task_data_renamed(
            migrate_from="0025_base_manager_name",
            migrate_to="0031_rename_more_arch_all_build_architecture_and_suite",
            old_work_requests=[
                (
                    TaskTypes.WORKFLOW,
                    "foo",
                    {"publish_target": "bookworm@debian:suite"},
                    {},
                    {"publish_target": "bookworm@debian:suite"},
                ),
                (
                    TaskTypes.WORKFLOW,
                    "debian_pipeline",
                    {"publish_target": "bookworm@debian:suite"},
                    {},
                    {"publish_target": "bookworm@debian:suite"},
                ),
                (
                    TaskTypes.WORKFLOW,
                    "debian_pipeline",
                    {"vendor": "debian"},
                    {},
                    {"vendor": "debian"},
                ),
            ],
            new_work_requests=[
                (
                    TaskTypes.WORKFLOW,
                    "foo",
                    {"publish_target": "bookworm@debian:suite"},
                    {},
                    {"publish_target": "bookworm@debian:suite"},
                ),
                (
                    TaskTypes.WORKFLOW,
                    "debian_pipeline",
                    {"suite": "bookworm@debian:suite"},
                    {},
                    {"suite": "bookworm@debian:suite"},
                ),
                (
                    TaskTypes.WORKFLOW,
                    "debian_pipeline",
                    {"vendor": "debian"},
                    {},
                    {"vendor": "debian"},
                ),
            ],
        )
        self.assert_workflow_template_task_data_renamed(
            migrate_from="0025_base_manager_name",
            migrate_to="0026_rename_publish_target_to_suite",
            old_workflow_templates=[
                ("foo", {"publish_target": "bookworm@debian:suite"}),
                (
                    "debian_pipeline",
                    {"publish_target": "bookworm@debian:suite"},
                ),
                ("debian_pipeline", {"vendor": "debian"}),
            ],
            new_workflow_templates=[
                ("foo", {"publish_target": "bookworm@debian:suite"}),
                ("debian_pipeline", {"suite": "bookworm@debian:suite"}),
                ("debian_pipeline", {"vendor": "debian"}),
            ],
        )

    def test_rename_worker_static_native_architecture(self) -> None:
        migrate_from = [("db", "0026_rename_publish_target_to_suite")]
        migrate_to = [("db", "0027_rename_worker_static_native_architecture")]
        old_workers = [
            ("worker-1", {}),
            (
                "worker-2",
                {
                    "system:host_architecture": "amd64",
                    "system:architectures": ["amd64", "i386"],
                },
            ),
        ]
        new_workers = [
            ("worker-1", {}),
            (
                "worker-2",
                {
                    "system:native_architecture": "amd64",
                    "system:architectures": ["amd64", "i386"],
                },
            ),
        ]

        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(migrate_from)
        old_apps = executor.loader.project_state(migrate_from).apps
        for name, static_metadata in old_workers:
            old_apps.get_model("db", "Worker").objects.create(
                name=name,
                registered_at=timezone.now(),
                token=old_apps.get_model("db", "Token").objects.create(
                    hash=hashlib.sha256(name.encode()).hexdigest()
                ),
                static_metadata=static_metadata,
            )

        executor.loader.build_graph()
        executor.migrate(migrate_to)
        new_apps = executor.loader.project_state(migrate_to).apps

        self.assertCountEqual(
            new_apps.get_model("db", "Worker").objects.values_list(
                "name", "static_metadata"
            ),
            new_workers,
        )

        executor.loader.build_graph()
        executor.migrate(migrate_from)

        self.assertCountEqual(
            old_apps.get_model("db", "Worker").objects.values_list(
                "name", "static_metadata"
            ),
            old_workers,
        )

    def test_rename_arch_all_build_architecture(self) -> None:
        affected_workflows = {
            "autopkgtest",
            "blhc",
            "debdiff",
            "debian_pipeline",
            "lintian",
            "package_upload",
            "piuparts",
            "qa",
            "reverse_dependencies_autopkgtest",
            "sbuild",
        }
        self.assert_work_request_all_task_data_renamed(
            migrate_from="0027_rename_worker_static_native_architecture",
            migrate_to="0031_rename_more_arch_all_build_architecture_and_suite",
            old_work_requests=[
                (
                    TaskTypes.WORKFLOW,
                    "foo",
                    {"arch_all_host_architecture": "amd64"},
                    {},
                    {"arch_all_host_architecture": "amd64"},
                ),
                *(
                    (
                        TaskTypes.WORKFLOW,
                        workflow,
                        {
                            "vendor": "debian",
                            "arch_all_host_architecture": "amd64",
                        },
                        {},
                        {
                            "vendor": "debian",
                            "arch_all_host_architecture": "amd64",
                        },
                    )
                    for workflow in affected_workflows
                ),
            ],
            new_work_requests=[
                (
                    TaskTypes.WORKFLOW,
                    "foo",
                    {"arch_all_host_architecture": "amd64"},
                    {},
                    {"arch_all_host_architecture": "amd64"},
                ),
                *(
                    (
                        TaskTypes.WORKFLOW,
                        workflow,
                        {
                            "vendor": "debian",
                            "arch_all_build_architecture": "amd64",
                        },
                        {},
                        {
                            "vendor": "debian",
                            "arch_all_build_architecture": "amd64",
                        },
                    )
                    for workflow in affected_workflows
                ),
            ],
        )
        self.assert_workflow_template_task_data_renamed(
            migrate_from="0027_rename_worker_static_native_architecture",
            migrate_to="0028_rename_arch_all_build_architecture",
            old_workflow_templates=[
                ("foo", {"arch_all_host_architecture": "amd64"}),
                *(
                    (
                        workflow,
                        {
                            "vendor": "debian",
                            "arch_all_host_architecture": "amd64",
                        },
                    )
                    for workflow in affected_workflows
                ),
            ],
            new_workflow_templates=[
                ("foo", {"arch_all_host_architecture": "amd64"}),
                *(
                    (
                        workflow,
                        {
                            "vendor": "debian",
                            "arch_all_build_architecture": "amd64",
                        },
                    )
                    for workflow in affected_workflows
                ),
            ],
        )

    def test_rename_build_architecture(self) -> None:
        affected_tasks = {
            "autopkgtest",
            "lintian",
            "blhc",
            "debdiff",
            "piuparts",
            "sbuild",
            "extractforsigning",
            "assemblesignedsource",
            "makesourcepackageupload",
        }
        self.assert_work_request_all_task_data_renamed(
            migrate_from="0028_rename_arch_all_build_architecture",
            migrate_to="0030_rename_more_build_architecture",
            old_work_requests=[
                (
                    TaskTypes.WORKER,
                    "foo",
                    {"host_architecture": "amd64"},
                    {"host_architecture": "amd64"},
                    {"host_architecture": "amd64"},
                ),
                *(
                    (
                        TaskTypes.WORKER,
                        task,
                        {"backend": "unshare", "host_architecture": "amd64"},
                        {"backend": "unshare", "host_architecture": "amd64"},
                        {"backend": "unshare", "host_architecture": "amd64"},
                    )
                    for task in affected_tasks
                ),
            ],
            new_work_requests=[
                (
                    TaskTypes.WORKER,
                    "foo",
                    {"host_architecture": "amd64"},
                    {"host_architecture": "amd64"},
                    {"host_architecture": "amd64"},
                ),
                *(
                    (
                        TaskTypes.WORKER,
                        task,
                        {"backend": "unshare", "build_architecture": "amd64"},
                        {"backend": "unshare", "build_architecture": "amd64"},
                        {"backend": "unshare", "build_architecture": "amd64"},
                    )
                    for task in affected_tasks
                ),
            ],
        )

    def test_add_runtime_parameters(self) -> None:
        """Runtime parameters are generated on workflow templates."""
        migrate_from = [
            ("db", "0031_rename_more_arch_all_build_architecture_and_suite")
        ]
        migrate_to = [
            ("db", "0034_workflow_template_configure_runtime_parameters")
        ]

        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(migrate_from)

        old_apps = executor.loader.project_state(migrate_from).apps
        workspace = old_apps.get_model("db", "Workspace").objects.get(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        OldWorkflowTemplate = old_apps.get_model("db", "WorkflowTemplate")

        old_template = OldWorkflowTemplate.objects.create(
            name="test",
            workspace=workspace,
            task_name="create_experiment_workspace",
            task_data={
                "public": True,
                "workflow_template_names": ["upload-to-foo"],
                "expiration_delay": 60,
            },
            priority=0,
        )

        executor.loader.build_graph()
        executor.migrate(migrate_to)

        new_apps = executor.loader.project_state(migrate_to).apps
        NewWorkflowTemplate = new_apps.get_model("db", "WorkflowTemplate")

        new_template = NewWorkflowTemplate.objects.get(pk=old_template.pk)
        self.assertEqual(new_template.static_parameters, old_template.task_data)
        self.assertEqual(
            new_template.runtime_parameters,
            {
                "experiment_name": "any",
                "owner_group": "any",
                "task_configuration": "any",
            },
        )

        executor.loader.build_graph()
        executor.migrate(migrate_from)

        new_old_template = OldWorkflowTemplate.objects.get(pk=old_template.pk)
        self.assertEqual(new_old_template.task_data, old_template.task_data)

    def test_rename_create_child_workspace(self) -> None:
        self.assert_work_request_all_task_data_renamed(
            migrate_from="0035_alter_workrequest_status",
            migrate_to="0036_rename_create_child_workspace",
            old_work_requests=[
                (
                    TaskTypes.SERVER,
                    "create_experiment_workspace",
                    {"experiment_name": "test"},
                    {},
                    {"experiment_name": "test"},
                ),
                (
                    TaskTypes.SERVER,
                    "create_experiment_workspace",
                    {"experiment_name": "test2"},
                    {},
                    None,
                ),
                (TaskTypes.SERVER, "servernoop", {}, {}, {}),
                (
                    TaskTypes.WORKFLOW,
                    "create_experiment_workspace",
                    {"experiment_name": "test"},
                    {},
                    {"experiment_name": "test"},
                ),
                (
                    TaskTypes.WORKFLOW,
                    "create_experiment_workspace",
                    {"experiment_name": "test2"},
                    {},
                    None,
                ),
                (TaskTypes.WORKFLOW, "noop", {}, {}, {}),
            ],
            new_work_requests=[
                (
                    TaskTypes.SERVER,
                    "create_child_workspace",
                    {
                        "prefix": settings.DEBUSINE_DEFAULT_WORKSPACE,
                        "suffix": "test",
                    },
                    {},
                    {
                        "prefix": settings.DEBUSINE_DEFAULT_WORKSPACE,
                        "suffix": "test",
                    },
                ),
                (
                    TaskTypes.SERVER,
                    "create_child_workspace",
                    {
                        "prefix": settings.DEBUSINE_DEFAULT_WORKSPACE,
                        "suffix": "test2",
                    },
                    {},
                    None,
                ),
                (TaskTypes.SERVER, "servernoop", {}, {}, {}),
                (
                    TaskTypes.WORKFLOW,
                    "create_child_workspace",
                    {"suffix": "test"},
                    {},
                    {"suffix": "test"},
                ),
                (
                    TaskTypes.WORKFLOW,
                    "create_child_workspace",
                    {"suffix": "test2"},
                    {},
                    None,
                ),
                (TaskTypes.WORKFLOW, "noop", {}, {}, {}),
            ],
        )
        self.assert_workflow_template_parameters_renamed(
            migrate_from="0035_alter_workrequest_status",
            migrate_to="0036_rename_create_child_workspace",
            old_workflow_templates=[
                ("foo", {}, {"experiment_name": RuntimeParameter.ANY}),
                (
                    "create_experiment_workspace",
                    {},
                    {"experiment_name": RuntimeParameter.ANY},
                ),
                (
                    "create_experiment_workspace",
                    {"experiment_name": "test"},
                    {},
                ),
            ],
            new_workflow_templates=[
                ("foo", {}, {"experiment_name": RuntimeParameter.ANY}),
                (
                    "create_child_workspace",
                    {},
                    {"suffix": RuntimeParameter.ANY},
                ),
                ("create_child_workspace", {"suffix": "test"}, {}),
            ],
        )

    def test_copy_build_arch_to_tag(self) -> None:
        BUILD_ARCH_PREFIX = "worker:build-arch:"

        migrate_from = [("db", "0045_alter_workrequest_unblock_strategy")]
        migrate_to = [("db", "0046_copy_build_arch_to_tag")]

        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(migrate_from)

        old_apps = executor.loader.project_state(migrate_from).apps
        OldWorkspace = old_apps.get_model("db", "Workspace")
        old_workspace = OldWorkspace.objects.get(
            name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        old_user = self.get_test_user_for_apps(old_apps)
        OldWorkRequest = old_apps.get_model("db", "WorkRequest")

        old = {
            # A task without build architecture (like noop)
            "noop": OldWorkRequest.objects.create(
                workspace=old_workspace,
                created_by=old_user,
                task_type=TaskTypes.WORKER,
                task_name="noop",
                task_data={},
            ),
            # A task that uses build_architecture (like sbuild)
            "sbuild": OldWorkRequest.objects.create(
                workspace=old_workspace,
                created_by=old_user,
                task_type=TaskTypes.WORKER,
                task_name="sbuild",
                task_data={"build_architecture": "risc64"},
            ),
            # A sbuild-like task with an undefined build_architecture
            "sbuild_none": OldWorkRequest.objects.create(
                workspace=old_workspace,
                created_by=old_user,
                task_type=TaskTypes.WORKER,
                task_name="sbuild",
                task_data={"build_architecture": None},
            ),
            # A task that uses bootstrap_options.architecture (like mmdebstrap)
            "mmdebstrap": OldWorkRequest.objects.create(
                workspace=old_workspace,
                created_by=old_user,
                task_type=TaskTypes.WORKER,
                task_name="mmdebstrap",
                task_data={"bootstrap_options": {"architecture": "arm64"}},
            ),
            # A mmdebstrap-like task with an undefined build architecture
            "mmdebstrap_none": OldWorkRequest.objects.create(
                workspace=old_workspace,
                created_by=old_user,
                task_type=TaskTypes.WORKER,
                task_name="mmdebstrap",
                task_data={"bootstrap_options": {"architecture": None}},
            ),
            # A task that already has some task-required tags
            "sbuild_with_tags": OldWorkRequest.objects.create(
                workspace=old_workspace,
                created_by=old_user,
                task_type=TaskTypes.WORKER,
                task_name="sbuild",
                task_data={"build_architecture": "i386"},
                scheduler_tags_required=["existing_tag"],
            ),
            # A task that already has some task-required tags
            "sbuild_with_arch_tags": OldWorkRequest.objects.create(
                workspace=old_workspace,
                created_by=old_user,
                task_type=TaskTypes.WORKER,
                task_name="sbuild",
                task_data={"build_architecture": "ppc64el"},
                scheduler_tags_required=[
                    "existing_tag",
                    BUILD_ARCH_PREFIX + "ppc64el",
                ],
            ),
            # A task that already has some task-required tags
            "mmdebstrap_with_arch_tags": OldWorkRequest.objects.create(
                workspace=old_workspace,
                created_by=old_user,
                task_type=TaskTypes.WORKER,
                task_name="sbuild",
                task_data={"bootstrap_options": {"architecture": "ppc64el"}},
                scheduler_tags_required=[
                    "existing_tag",
                    BUILD_ARCH_PREFIX + "ppc64el",
                ],
            ),
        }

        expected_tags = {
            "noop": [],
            "sbuild": [BUILD_ARCH_PREFIX + "risc64"],
            "sbuild_none": [],
            "mmdebstrap": [BUILD_ARCH_PREFIX + "arm64"],
            "mmdebstrap_none": [],
            "sbuild_with_tags": ["existing_tag", BUILD_ARCH_PREFIX + "i386"],
            "sbuild_with_arch_tags": [
                "existing_tag",
                BUILD_ARCH_PREFIX + "ppc64el",
            ],
            "mmdebstrap_with_arch_tags": [
                "existing_tag",
                BUILD_ARCH_PREFIX + "ppc64el",
            ],
        }

        executor.loader.build_graph()
        executor.migrate(migrate_to)

        new_apps = executor.loader.project_state(migrate_to).apps
        NewWorkRequest = new_apps.get_model("db", "WorkRequest")

        for name in old:
            with self.subTest(name=name):
                new = NewWorkRequest.objects.get(pk=old[name].pk)
                self.assertEqual(
                    new.scheduler_tags_required, expected_tags[name]
                )
        executor.loader.build_graph()
        executor.migrate(migrate_from)

        # Going backwards does not remove tags, as it may remove tags that were
        # preexisting before the migration
        for name in old:
            with self.subTest(name=name):
                new_old = OldWorkRequest.objects.get(pk=old[name].pk)
                self.assertEqual(
                    new_old.scheduler_tags_required, expected_tags[name]
                )

    def test_set_token_workers(self) -> None:
        migrate_from = [("db", "0048_mirror_worker_token_ids")]
        migrate_to = [("db", "0049_token_worker_data")]

        executor = MigrationExecutor(
            connection, progress_callback=_migrate_check_constraints
        )
        executor.migrate(migrate_from)

        old_apps = executor.loader.project_state(migrate_from).apps
        OldToken = old_apps.get_model("db", "Token")
        OldWorker = old_apps.get_model("db", "Worker")

        old_tokens = [
            OldToken.objects.create(
                hash=hashlib.sha256(str(i).encode()).hexdigest(), enabled=True
            )
            for i in range(6)
        ]
        old_workers = {
            worker.name: worker
            for worker in (
                OldWorker.objects.create(
                    name="celery-worker-without-token",
                    registered_at=timezone.now(),
                    worker_type=WorkerType.CELERY,
                ),
                OldWorker.objects.create(
                    name="worker-with-activation-token-1",
                    registered_at=timezone.now(),
                    activation_token=old_tokens[0],
                ),
                OldWorker.objects.create(
                    name="worker-with-activation-token-2",
                    registered_at=timezone.now(),
                    activation_token=old_tokens[1],
                ),
                OldWorker.objects.create(
                    name="worker-with-token-1",
                    registered_at=timezone.now(),
                    token=old_tokens[2],
                ),
                OldWorker.objects.create(
                    name="worker-with-token-2",
                    registered_at=timezone.now(),
                    token=old_tokens[3],
                ),
            )
        }
        old_user = self.get_test_user_for_apps(old_apps)
        old_tokens[4].user = old_user
        old_tokens[4].save()

        expected_tokens = [
            (
                old_tokens[0].id,
                _TokenTypes.WORKER_ACTIVATION,
                old_workers["worker-with-activation-token-1"].id,
                None,
            ),
            (
                old_tokens[1].id,
                _TokenTypes.WORKER_ACTIVATION,
                old_workers["worker-with-activation-token-2"].id,
                None,
            ),
            (
                old_tokens[2].id,
                _TokenTypes.WORKER,
                None,
                old_workers["worker-with-token-1"].id,
            ),
            (
                old_tokens[3].id,
                _TokenTypes.WORKER,
                None,
                old_workers["worker-with-token-2"].id,
            ),
            (old_tokens[4].id, _TokenTypes.USER, None, None),
        ]

        executor.loader.build_graph()
        executor.migrate(migrate_to)

        new_apps = executor.loader.project_state(migrate_to).apps
        NewToken = new_apps.get_model("db", "Token")

        self.assertQuerySetEqual(
            (
                NewToken.objects.filter(
                    id__in={token.id for token in old_tokens}
                )
                .order_by("id")
                .values_list(
                    "id", "token_type", "_activating_worker_id", "_worker_id"
                )
            ),
            expected_tokens,
        )
        self.assertFalse(NewToken.objects.filter(id=old_tokens[5].id).exists())

        executor.loader.build_graph()
        executor.migrate(migrate_from)

        self.assertQuerySetEqual(
            (
                OldToken.objects.filter(
                    id__in={token.id for token in old_tokens}
                )
                .order_by("id")
                .values_list(
                    "id", "token_type", "_activating_worker_id", "_worker_id"
                )
            ),
            [(token.id, "", None, None) for token in old_tokens[:-1]],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the piuparts workflow."""

import datetime as dt
from collections.abc import Sequence
from typing import ClassVar, override
from unittest import mock

from django.test import override_settings
from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianPiuparts,
    DebianSourcePackage,
    TaskTypes,
)
from debusine.client.models import LookupChildType
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Collection,
    CollectionItem,
    WorkRequest,
)
from debusine.db.models.work_requests import SkipWorkRequest
from debusine.server.workflows import WorkflowValidationError
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import (
    BaseWorkflowData,
    PiupartsWorkflowData,
    PiupartsWorkflowDynamicData,
    RegressionTrackingUpdateQAResults,
    SbuildWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.piuparts import PiupartsWorkflow
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestCase,
)
from debusine.server.workflows.tests.test_regression_tracking import (
    RegressionTrackingWorkflowTestMixin,
)
from debusine.tasks.models import (
    ActionSkipIfLookupResultChanged,
    ActionUpdateCollectionWithArtifacts,
    BackendType,
    BaseDynamicTaskData,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    OutputData,
    RegressionAnalysisStatus,
    SbuildData,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry


class PiupartsWorkflowTests(
    RegressionTrackingWorkflowTestMixin, WorkflowTestCase[PiupartsWorkflow]
):
    """Unit tests for :py:class:`PiupartsWorkflow`."""

    SAMPLE_TASK_DATA = {
        "vendor": "debian",
        "codename": "trixie",
    }

    source_artifact: ClassVar[Artifact]
    binary_artifact1: ClassVar[Artifact]
    binary_artifact2: ClassVar[Artifact]
    binary_artifact3: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        cls.binary_artifact1 = (
            cls.playground.create_minimal_binary_package_artifact("hello1")
        )
        cls.binary_artifact2 = (
            cls.playground.create_minimal_binary_package_artifact("hello2")
        )
        cls.binary_artifact3 = (
            cls.playground.create_minimal_binary_package_artifact("hello3")
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["source_artifact"] = self.source_artifact.pk
        self.task_data["binary_artifacts"] = [
            self.binary_artifact1.pk,
            self.binary_artifact2.pk,
            self.binary_artifact3.pk,
        ]

    def add_qa_result(
        self,
        qa_results: Collection,
        package: str,
        version: str,
        architecture: str,
        *,
        binary_packages: dict[str, str],
        distribution: str,
    ) -> CollectionItem:
        """Add a piuparts result to a ``debian:qa-results`` collection."""
        work_request = self.playground.create_worker_task(
            task_name="piuparts",
            result=WorkRequest.Results.SUCCESS,
            validate=False,
            use_data=False,
        )
        return qa_results.manager.add_artifact(
            self.playground.create_artifact(
                category=ArtifactCategory.PIUPARTS,
                data=DebianPiuparts(
                    piuparts_version="1.6.0",
                    binary_packages=binary_packages,
                    architecture=architecture,
                    distribution=distribution,
                ),
            )[0],
            user=self.playground.get_default_user(),
            variables={
                "package": package,
                "version": version,
                "architecture": architecture,
                "timestamp": int(timezone.now().timestamp()),
                "work_request_id": work_request.id,
            },
        )

    def orchestrate(
        self,
        *,
        task_data: PiupartsWorkflowData,
        source_artifact: Artifact,
        architectures: Sequence[str],
        parent: WorkRequest | None = None,
        pipeline_task_name: str = "examplepipeline",
    ) -> WorkRequest:
        """Create and orchestrate a PiupartsWorkflow."""

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """
            Pipeline workflow that provides a promise (e.g. by Sbuild).

            Piuparts will use the promise.
            """

            TASK_NAME = pipeline_task_name

            @override
            def populate(self_) -> None:
                source_data = source_artifact.create_data()
                assert isinstance(source_data, DebianSourcePackage)

                sbuild, _ = self_.work_request_ensure_child_workflow(
                    task_name="sbuild",
                    task_data=SbuildWorkflowData(
                        input=SbuildInput(source_artifact=source_artifact.id),
                        target_distribution="debian:sid",
                        architectures=list(architectures),
                    ),
                    workflow_data=WorkRequestWorkflowData(
                        display_name="sbuild", step="sbuild"
                    ),
                )
                self.playground.advance_work_request(sbuild, mark_running=True)
                for architecture in architectures:
                    child = sbuild.create_child_worker(
                        task_name="sbuild",
                        task_data=SbuildData(
                            input=SbuildInput(
                                source_artifact=source_artifact.id
                            ),
                            build_architecture=architecture,
                            environment="debian/match:codename=sid",
                        ),
                    )
                    self_.provides_artifact(
                        child,
                        ArtifactCategory.BINARY_PACKAGE,
                        f"build-{architecture}",
                        data={
                            "source_package_name": source_data.name,
                            "architecture": architecture,
                        },
                    )
                sbuild.unblock_workflow_children()

                piuparts, _ = self_.work_request_ensure_child_workflow(
                    task_name="piuparts",
                    task_data=task_data,
                    workflow_data=WorkRequestWorkflowData(
                        display_name="piuparts", step="piuparts"
                    ),
                )
                self_.orchestrate_child(piuparts)

        root = self.playground.create_workflow(
            task_name=pipeline_task_name, parent=parent
        )
        self.assertTrue(
            self.playground.orchestrate_workflow(root, use_data=False)
        )

        return root

    def test_validate_input(self) -> None:
        """validate_input passes a valid case."""
        self.task.validate_input()

    def test_validate_input_bad_qa_suite(self) -> None:
        """validate_input raises errors in looking up a suite."""
        self.task_data["qa_suite"] = "nonexistent@debian:suite"
        self.validate = False
        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:suite' does not exist or is hidden",
        ):
            self.task.validate_input()

    def test_validate_input_bad_reference_qa_results(self) -> None:
        """validate_input raises errors in looking up reference QA results."""
        self.task_data["reference_qa_results"] = "nonexistent@debian:qa-results"
        self.validate = False
        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:qa-results' does not exist or is hidden",
        ):
            self.task.validate_input()

    def test_has_current_reference_qa_result_no_match(self) -> None:
        """_has_current_reference_qa_result: no matching result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(
            sid_qa_results,
            "other",
            "1.0-1",
            "amd64",
            binary_packages={"other": "1.0-1"},
            distribution="debian:sid",
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello", srcpkg_version="1.0-1"
            )
        )

        wr = self.playground.create_workflow(
            task_name="piuparts",
            task_data=PiupartsWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result("amd64")
        )

    def test_has_current_reference_qa_result_different_version(self) -> None:
        """_has_current_reference_qa_result: result for different version."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(
            sid_qa_results,
            "hello",
            "1.0-1",
            "amd64",
            binary_packages={"hello": "1.0-1"},
            distribution="debian:sid",
        )
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello", srcpkg_version="1.0-2"
            )
        )

        wr = self.playground.create_workflow(
            task_name="piuparts",
            task_data=PiupartsWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result("amd64")
        )

    def test_has_current_reference_qa_result_current(self) -> None:
        """_has_current_reference_qa_result: current result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(
            sid_qa_results,
            "hello",
            "1.0-1",
            "amd64",
            binary_packages={"hello": "1.0-1"},
            distribution="debian:sid",
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello", srcpkg_version="1.0-1"
            )
        )

        wr = self.playground.create_workflow(
            task_name="piuparts",
            task_data=PiupartsWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.assertTrue(
            self.get_workflow(wr)._has_current_reference_qa_result("amd64")
        )

    def _create_package_artifacts(self) -> tuple[Artifact, dict[str, Artifact]]:
        """Return source artifact and dictionary with binary artifacts."""
        return (
            self.playground.create_source_artifact(
                name="hello", version="1.0.0"
            ),
            {
                "all": self.playground.create_minimal_binary_package_artifact(
                    architecture="all"
                ),
                "amd64": self.playground.create_minimal_binary_package_artifact(
                    architecture="amd64"
                ),
                "i386": self.playground.create_minimal_binary_package_artifact(
                    architecture="i386"
                ),
            },
        )

    def _create_debian_environments(
        self, *, codename: str, architectures: list[str]
    ) -> None:
        for architecture in architectures:
            self.playground.create_debian_environment(
                architecture=architecture,
                codename=codename,
            )

    @preserve_db_task_registry()
    def test_populate(self) -> None:
        """Test populate."""
        # Expected architectures as per intersection of architectures of
        # binary_artifacts and requested architectures
        self._create_debian_environments(
            codename="bookworm", architectures=["amd64", "i386"]
        )
        source, binaries = self._create_package_artifacts()

        o = self.playground.create_workflow(
            task_name="piuparts",
            task_data={
                "source_artifact": source.id,
                "binary_artifacts": [
                    binaries["all"].id,
                    binaries["amd64"].id,
                    binaries["i386"].id,
                ],
                "vendor": "debian",
                "codename": "bookworm",
                "extra_repositories": [
                    {
                        "url": "http://example.com/",
                        "suite": "bookworm",
                        "components": ["main"],
                    }
                ],
            },
        )
        self.get_workflow(o).populate()

        piuparts_wrs = WorkRequest.objects.filter(
            task_type=TaskTypes.WORKER,
            task_name="piuparts",
            parent=o,
        )
        self.assertEqual(piuparts_wrs.count(), 2)

        for architecture in ("amd64", "i386"):
            with self.subTest(architecture=architecture):
                piuparts_wr = piuparts_wrs.get(
                    task_data__input__binary_artifacts__contains=(
                        binaries["amd64"].id
                    )
                )
                self.assertEqual(
                    piuparts_wr.task_data,
                    {
                        "backend": BackendType.UNSHARE,
                        "input": {
                            "binary_artifacts": sorted(
                                [binaries["all"].id, binaries["amd64"].id]
                            )
                        },
                        "build_architecture": "amd64",
                        "base_tgz": "debian/match:codename=bookworm",
                        "environment": "debian/match:codename=bookworm",
                        "extra_repositories": [
                            {
                                "url": "http://example.com/",
                                "suite": "bookworm",
                                "components": ["main"],
                            }
                        ],
                    },
                )

                self.assert_work_request_event_reactions(
                    piuparts_wr,
                    on_success=[
                        ActionUpdateCollectionWithArtifacts(
                            artifact_filters={
                                "category": ArtifactCategory.PIUPARTS
                            },
                            collection="internal@collections",
                            name_template="piuparts-amd64",
                        )
                    ],
                )

    def test_populate_architecture_only_i386(self) -> None:
        """Test "architectures=["i386"]: do not build on amd64."""
        self._create_debian_environments(
            codename="trixie", architectures=["amd64", "i386"]
        )
        source, binaries = self._create_package_artifacts()

        o = self.playground.create_workflow(
            task_name="piuparts",
            task_data={
                "source_artifact": source.id,
                "binary_artifacts": [
                    binaries["all"].id,
                    binaries["amd64"].id,
                    binaries["i386"].id,
                ],
                "vendor": "debian",
                "codename": "trixie",
                "architectures": ["i386"],
            },
        )
        self.get_workflow(o).populate()

        piuparts_wrs = WorkRequest.objects.filter(
            task_type=TaskTypes.WORKER,
            task_name="piuparts",
            parent=o,
        )

        self.assertEqual(piuparts_wrs.count(), 1)

        piuparts_i386_wr = piuparts_wrs.get(
            task_data__input__binary_artifacts__contains=(
                binaries["all"].id,
                binaries["i386"].id,
            ),
            task_data__build_architecture="i386",
        )
        self.assertEqual(
            piuparts_i386_wr.task_data,
            {
                "backend": BackendType.UNSHARE,
                "input": {
                    "binary_artifacts": sorted(
                        [
                            binaries["all"].id,
                            binaries["i386"].id,
                        ]
                    )
                },
                "build_architecture": "i386",
                "base_tgz": "debian/match:codename=trixie",
                "environment": "debian/match:codename=trixie",
                "extra_repositories": None,
            },
        )

    def test_populate_experimental(self) -> None:
        """The workflow handles overlay distributions."""
        self.playground.create_debian_environment(
            codename="experimental", variant="piuparts"
        )
        self.task_data["binary_artifacts"] = [self.binary_artifact1.pk]
        self.task_data["codename"] = "experimental"
        self.task.populate()
        children = list(self.work_request.children.all())
        self.assertEqual(len(children), 1)
        child = children[0]
        self.assertEqual(len(child.task_data["extra_repositories"]), 1)
        repo = child.task_data["extra_repositories"][0]
        self.assertEqual(repo["suite"], "experimental")

    def test_populate_uploads(self) -> None:
        """The workflow accepts debian:upload artifacts in all inputs."""
        artifacts = self.playground.create_upload_artifacts(
            binaries=[
                ("hello", "amd64"),
                ("hello", "i386"),
                ("hello-doc", "all"),
            ]
        )
        architectures = ["amd64", "i386"]
        codename = "sid"

        for architecture in architectures:
            self.playground.create_debian_environment(
                codename=codename,
                architecture=architecture,
            )

        wr = self.playground.create_workflow(
            task_name="piuparts",
            task_data=PiupartsWorkflowData(
                source_artifact=artifacts.upload.id,
                binary_artifacts=LookupMultiple((artifacts.upload.id,)),
                architectures=architectures,
                vendor="debian",
                codename=codename,
            ),
        )

        self.get_workflow(wr).populate()
        children = list(wr.children.order_by("id").all())
        self.assertEqual(len(children), len(architectures))
        for architecture, child in zip(architectures, children):
            with self.subTest(architecture=architecture):
                self.assertEqual(
                    child.task_data["build_architecture"], architecture
                )
                self.assertEqual(
                    child.task_data["input"],
                    {
                        "binary_artifacts": sorted(
                            f"{artifact.id}@artifacts"
                            for artifact in artifacts.binaries
                            if artifact.data["deb_fields"]["Architecture"]
                            in (architecture, "all")
                        ),
                    },
                )

    def test_populate_has_current_reference_qa_result(self) -> None:
        """The workflow does nothing with a current reference QA result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(
            sid_qa_results,
            "hello",
            "1.0-1",
            "amd64",
            binary_packages={"hello": "1.0-1"},
            distribution="debian:sid",
        )

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        wr = self.playground.create_workflow(
            task_name="piuparts",
            task_data=PiupartsWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.get_workflow(wr).populate()

        self.assertQuerySetEqual(wr.children.all(), [])

    def test_populate_has_current_reference_qa_result_force(self) -> None:
        """The workflow can be forced despite a current reference QA result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(
            sid_qa_results,
            "hello",
            "1.0-1",
            "amd64",
            binary_packages={"hello": "1.0-1"},
            distribution="debian:sid",
        )

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        wr = self.playground.create_workflow(
            task_name="piuparts",
            task_data=PiupartsWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.FORCE,
            ),
        )

        self.get_workflow(wr).populate()

        self.assertTrue(
            wr.children.filter(
                task_type=TaskTypes.WORKER, task_name="piuparts"
            ).exists()
        )

    def test_populate_no_previous_reference_qa_result(self) -> None:
        """The workflow produces a reference QA result if there is none."""
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        self.playground.create_collection("sid", CollectionCategory.QA_RESULTS)

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        source_item = sid_suite.manager.add_artifact(
            self.source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        wr = self.playground.create_workflow(
            task_name="piuparts",
            task_data=PiupartsWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.get_workflow(wr).populate()

        [child] = wr.children.all()
        self.assertEqual(child.status, WorkRequest.Statuses.BLOCKED)
        self.assertEqual(child.task_type, TaskTypes.WORKER)
        self.assertEqual(child.task_name, "piuparts")
        self.assertEqual(
            child.task_data,
            {
                "backend": BackendType.UNSHARE,
                "base_tgz": "debian/match:codename=sid",
                "build_architecture": "amd64",
                "environment": "debian/match:codename=sid",
                "extra_repositories": None,
                "input": {"binary_artifacts": [binary_artifact.id]},
            },
        )
        provide_action = ActionUpdateCollectionWithArtifacts(
            artifact_filters={"category": ArtifactCategory.PIUPARTS},
            collection="internal@collections",
            name_template="reference-qa-result|piuparts-amd64",
        )
        skip_action = ActionSkipIfLookupResultChanged(
            lookup=(
                f"sid@{CollectionCategory.QA_RESULTS}/"
                f"latest:piuparts_hello_amd64"
            ),
            collection_item_id=None,
            promise_name=None,
        )
        qa_result_action = ActionUpdateCollectionWithArtifacts(
            artifact_filters={"category": ArtifactCategory.PIUPARTS},
            collection=f"sid@{CollectionCategory.QA_RESULTS}",
            variables={
                "package": "hello",
                "version": "1.0-1",
                "architecture": "amd64",
                "timestamp": int(source_item.created_at.timestamp()),
                "work_request_id": child.id,
            },
        )
        self.assert_work_request_event_reactions(
            child,
            on_assignment=[skip_action],
            on_failure=[qa_result_action],
            on_success=[provide_action, qa_result_action],
        )

        # Completing the work request stores the QA result.
        result, _ = self.playground.create_artifact(
            category=ArtifactCategory.PIUPARTS,
            data=DebianPiuparts(
                piuparts_version="1.6.0",
                binary_packages={"hello": "1.0-1"},
                architecture="amd64",
                distribution="debian:sid",
            ),
            work_request=child,
        )
        self.playground.advance_work_request(
            child, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/"
                f"latest:piuparts_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).artifact,
            result,
        )

        # Population is idempotent, even if qa_suite changes.
        with (
            mock.patch.object(
                PiupartsWorkflow,
                "_has_current_reference_qa_result",
                return_value=False,
            ),
            mock.patch.object(
                PiupartsWorkflow,
                "qa_suite_changed",
                mock.PropertyMock(
                    return_value=timezone.now() + dt.timedelta(hours=1)
                ),
            ),
        ):
            self.get_workflow(wr).populate()
        self.assertEqual(wr.children.count(), 1)
        child.refresh_from_db()
        self.assert_work_request_event_reactions(
            child,
            on_assignment=[skip_action],
            on_failure=[qa_result_action],
            on_success=[provide_action, qa_result_action],
        )

    def test_populate_reference_qa_result_backs_off(self) -> None:
        """Reference tasks are skipped if another workflow got there first."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.playground.create_debian_environment(
            codename="sid", variant="piuparts"
        )

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        wr = self.playground.create_workflow(
            task_name="piuparts",
            task_data=PiupartsWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.get_workflow(wr).populate()

        racing_qa_result = self.add_qa_result(
            sid_qa_results,
            "hello",
            "1.0-1",
            "amd64",
            binary_packages={"hello": "1.0-1"},
            distribution="debian:sid",
        )
        [child] = wr.children.all()
        self.playground.advance_work_request(child, mark_pending=True)

        with self.assertRaises(SkipWorkRequest):
            child.assign_worker(self.playground.create_worker())

        self.assertEqual(child.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(child.result, WorkRequest.Results.SKIPPED)
        self.assertEqual(
            child.output_data,
            OutputData(
                skip_reason=(
                    f"Result of lookup "
                    f"'sid@{CollectionCategory.QA_RESULTS}/"
                    f"latest:piuparts_hello_amd64' changed"
                )
            ),
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/"
                f"latest:piuparts_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).collection_item,
            racing_qa_result,
        )

    def test_populate_reference_qa_result_uses_build_architecture(self) -> None:
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        # Add a spurious QA result for "all" to make sure it isn't considered.
        self.add_qa_result(
            sid_qa_results,
            "hello",
            "1.0-1",
            "all",
            binary_packages={"hello": "1.0-1"},
            distribution="debian:sid",
        )

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="all",
            )
        )
        # At least one debian_environment must exist to create the
        # Piuparts work request for the "all" package
        self.playground.create_debian_environment(
            architecture="amd64",
            codename="sid",
        )

        source_item = sid_suite.manager.add_artifact(
            self.source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        wr = self.playground.create_workflow(
            task_name="piuparts",
            task_data=PiupartsWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
            ),
        )

        self.get_workflow(wr).populate()

        [child] = wr.children.all()
        self.assertEqual(child.status, WorkRequest.Statuses.BLOCKED)
        self.assertEqual(child.task_type, TaskTypes.WORKER)
        self.assertEqual(child.task_name, "piuparts")
        self.assertEqual(
            child.task_data,
            {
                "backend": BackendType.UNSHARE,
                "base_tgz": "debian/match:codename=sid",
                "build_architecture": "amd64",
                "environment": "debian/match:codename=sid",
                "extra_repositories": None,
                "input": {"binary_artifacts": [binary_artifact.id]},
            },
        )
        qa_result_action = ActionUpdateCollectionWithArtifacts(
            artifact_filters={"category": ArtifactCategory.PIUPARTS},
            collection=f"sid@{CollectionCategory.QA_RESULTS}",
            variables={
                "package": "hello",
                "version": "1.0-1",
                "architecture": "amd64",
                "timestamp": int(source_item.created_at.timestamp()),
                "work_request_id": child.id,
            },
        )
        self.assert_work_request_event_reactions(
            child,
            on_assignment=[
                ActionSkipIfLookupResultChanged(
                    lookup=(
                        f"sid@{CollectionCategory.QA_RESULTS}/"
                        f"latest:piuparts_hello_amd64"
                    ),
                    collection_item_id=None,
                    promise_name=None,
                )
            ],
            on_failure=[qa_result_action],
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": ArtifactCategory.PIUPARTS},
                    collection="internal@collections",
                    name_template="reference-qa-result|piuparts-amd64",
                ),
                qa_result_action,
            ],
        )

        # Completing the work request stores the QA result.
        result, _ = self.playground.create_artifact(
            category=ArtifactCategory.PIUPARTS,
            data=DebianPiuparts(
                piuparts_version="1.6.0",
                binary_packages={"hello": "1.0-1"},
                architecture="amd64",
                distribution="debian:sid",
            ),
            work_request=child,
        )
        self.playground.advance_work_request(
            child, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/"
                f"latest:piuparts_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).artifact,
            result,
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    @preserve_db_task_registry()
    def test_callback_regression_analysis(self) -> None:
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        self.playground.create_collection("sid", CollectionCategory.QA_RESULTS)
        architectures = ["amd64", "arm64", "i386", "ppc64el", "s390x"]

        reference_source_artifact = self.source_artifact
        reference_binary_artifacts = {
            arch: self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello", srcpkg_version="1.0-1", architecture=arch
            )
            for arch in architectures
            if arch != "ppc64el"
        }
        new_source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )
        new_binary_artifacts = {
            arch: self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello", srcpkg_version="1.0-2", architecture=arch
            )
            for arch in architectures
        }

        root = self.playground.create_workflow(mark_running=True)
        reference = self.orchestrate(
            task_data=PiupartsWorkflowData(
                prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                source_artifact=reference_source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        artifact.id
                        for artifact in reference_binary_artifacts.values()
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            source_artifact=reference_source_artifact,
            architectures=architectures,
            parent=root,
            pipeline_task_name="referencepipeline",
        )
        new = self.orchestrate(
            task_data=PiupartsWorkflowData(
                reference_prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                enable_regression_tracking=True,
                source_artifact=new_source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            source_artifact=new_source_artifact,
            architectures=architectures,
            parent=root,
        )

        # Unblock the piuparts tasks.
        for sbuild in reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=False
            )
        for sbuild in new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=False
            )

        reference_piuparts_workflow = reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="piuparts"
        )
        reference_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in reference_piuparts_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="piuparts"
            )
        }
        reference_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.PIUPARTS,
                data=DebianPiuparts(
                    piuparts_version="1.6.0",
                    binary_packages={"hello": "1.0-1"},
                    architecture=architecture,
                    distribution="debian:sid",
                ),
                work_request=reference_tasks[architecture],
            )[0]
            for architecture in architectures
            if architecture != "ppc64el"
        }
        for arch, artifact in reference_results.items():
            self.playground.create_artifact_relation(
                artifact,
                reference_binary_artifacts[arch],
                ArtifactRelation.Relations.RELATES_TO,
            )
        new_piuparts_workflow = new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="piuparts"
        )
        new_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in new_piuparts_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="piuparts"
            )
        }
        new_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.PIUPARTS,
                data=DebianPiuparts(
                    piuparts_version="1.6.0",
                    binary_packages={"hello": "1.0-1"},
                    architecture=architecture,
                    distribution="debian:sid",
                ),
                work_request=new_tasks[architecture],
            )[0]
            for architecture in architectures
        }
        for arch, artifact in new_results.items():
            self.playground.create_artifact_relation(
                artifact,
                new_binary_artifacts[arch],
                ArtifactRelation.Relations.RELATES_TO,
            )

        self.assertEqual(
            new_piuparts_workflow.output_data,
            OutputData(regression_analysis={}),
        )
        self.schedule_and_run_workflow_callback(new_piuparts_workflow)
        self.playground.advance_work_request(
            reference_tasks["amd64"],
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )
        self.schedule_and_run_workflow_callback(new_piuparts_workflow)
        self.playground.advance_work_request(
            new_tasks["amd64"],
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )

        self.schedule_and_run_workflow_callback(new_piuparts_workflow)

        expected_analysis: dict[str, tuple[RegressionAnalysisStatus, None]] = {
            "amd64": (RegressionAnalysisStatus.STABLE, None)
        }
        new_piuparts_workflow.refresh_from_db()
        self.assert_regression_analysis(
            new_piuparts_workflow,
            "1.0-1",
            reference_results,
            "1.0-2",
            new_results,
            expected_analysis,
        )

        self.assert_callback_status(
            new_piuparts_workflow, WorkRequest.Statuses.BLOCKED
        )

        for architecture, reference_result, new_result, status in (
            (
                "arm64",
                WorkRequest.Results.SUCCESS,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.REGRESSION,
            ),
            (
                "i386",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.STABLE,
            ),
            (
                "ppc64el",
                None,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.NO_RESULT,
            ),
            (
                "s390x",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.IMPROVEMENT,
            ),
        ):
            if reference_result is not None:
                self.playground.advance_work_request(
                    reference_tasks[architecture],
                    result=reference_result,
                    use_data=False,
                )
            self.playground.advance_work_request(
                new_tasks[architecture], result=new_result, use_data=False
            )
            expected_analysis[architecture] = (status, None)

        self.assert_callback_status(
            new_piuparts_workflow, WorkRequest.Statuses.PENDING
        )

        self.schedule_and_run_workflow_callback(new_piuparts_workflow)

        new_piuparts_workflow.refresh_from_db()
        self.assert_regression_analysis(
            new_piuparts_workflow,
            "1.0-1",
            reference_results,
            "1.0-2",
            new_results,
            expected_analysis,
        )
        self.assert_callback_status(
            new_piuparts_workflow, WorkRequest.Statuses.COMPLETED
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    @preserve_db_task_registry()
    def test_callback_regression_analysis_build_architecture(self) -> None:
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        self.playground.create_collection("sid", CollectionCategory.QA_RESULTS)

        reference_source_artifact = self.source_artifact
        reference_binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello", srcpkg_version="1.0-1", architecture="all"
            )
        )
        new_source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )
        new_binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello", srcpkg_version="1.0-2", architecture="all"
            )
        )
        # At least one debian_environment must exist to create the
        # Piuparts work request for the "all" package
        self.playground.create_debian_environment(
            architecture="amd64",
            codename="sid",
        )

        root = self.playground.create_workflow(mark_running=True)
        reference = self.orchestrate(
            task_data=PiupartsWorkflowData(
                prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                source_artifact=reference_source_artifact.id,
                binary_artifacts=LookupMultiple(
                    (reference_binary_artifact.id,)
                ),
                vendor="debian",
                codename="sid",
            ),
            source_artifact=reference_source_artifact,
            architectures=["all"],
            parent=root,
            pipeline_task_name="referencepipeline",
        )
        new = self.orchestrate(
            task_data=PiupartsWorkflowData(
                reference_prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                enable_regression_tracking=True,
                source_artifact=new_source_artifact.id,
                binary_artifacts=LookupMultiple(
                    ("internal@collections/name:build-all",)
                ),
                vendor="debian",
                codename="sid",
            ),
            source_artifact=new_source_artifact,
            architectures=["all"],
            parent=root,
        )

        # Unblock the piuparts tasks.
        for sbuild in reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=False
            )
        for sbuild in new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=False
            )

        reference_piuparts_workflow = reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="piuparts"
        )
        reference_task = reference_piuparts_workflow.children.get(
            task_type=TaskTypes.WORKER, task_name="piuparts"
        )
        reference_result, _ = self.playground.create_artifact(
            category=ArtifactCategory.PIUPARTS,
            data=DebianPiuparts(
                piuparts_version="1.6.0",
                binary_packages={"hello": "1.0-1"},
                architecture="amd64",
                distribution="debian:sid",
            ),
            work_request=reference_task,
        )
        self.playground.create_artifact_relation(
            reference_result,
            reference_binary_artifact,
            ArtifactRelation.Relations.RELATES_TO,
        )
        new_piuparts_workflow = new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="piuparts"
        )
        new_task = new_piuparts_workflow.children.get(
            task_type=TaskTypes.WORKER, task_name="piuparts"
        )
        new_result, _ = self.playground.create_artifact(
            category=ArtifactCategory.PIUPARTS,
            data=DebianPiuparts(
                piuparts_version="1.6.0",
                binary_packages={"hello": "1.0-1"},
                architecture="amd64",
                distribution="debian:sid",
            ),
            work_request=new_task,
        )
        self.playground.create_artifact_relation(
            new_result,
            new_binary_artifact,
            ArtifactRelation.Relations.RELATES_TO,
        )

        self.assertEqual(
            new_piuparts_workflow.output_data,
            OutputData(regression_analysis={}),
        )
        self.schedule_and_run_workflow_callback(new_piuparts_workflow)
        self.playground.advance_work_request(
            reference_task, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        self.schedule_and_run_workflow_callback(new_piuparts_workflow)
        self.playground.advance_work_request(
            new_task, result=WorkRequest.Results.SUCCESS, use_data=False
        )

        self.schedule_and_run_workflow_callback(new_piuparts_workflow)

        new_piuparts_workflow.refresh_from_db()
        self.assert_regression_analysis(
            new_piuparts_workflow,
            "1.0-1",
            {"amd64": reference_result},
            "1.0-2",
            {"amd64": new_result},
            {"amd64": (RegressionAnalysisStatus.STABLE, None)},
        )

        self.assert_callback_status(
            new_piuparts_workflow, WorkRequest.Statuses.COMPLETED
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data_binary_package(self) -> None:
        self.task_data["binary_artifacts"] = [self.binary_artifact1.pk]
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            PiupartsWorkflowDynamicData(
                subject="hello1",
                parameter_summary="hello1 in debian:trixie",
                source_artifact_id=self.source_artifact.id,
                binary_artifacts_ids=[self.binary_artifact1.id],
            ),
        )

    def test_compute_dynamic_data_multiple_source_packages(self) -> None:
        self.task_data["binary_artifacts"] = [
            self.binary_artifact1.id,
            self.binary_artifact2.id,
        ]
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            PiupartsWorkflowDynamicData(
                subject="hello1 hello2",
                parameter_summary="hello1 hello2 in debian:trixie",
                source_artifact_id=self.source_artifact.id,
                binary_artifacts_ids=[
                    self.binary_artifact1.id,
                    self.binary_artifact2.id,
                ],
            ),
        )

    def test_compute_dynamic_data_collection_raise_error(self) -> None:
        binary_artifact = (
            self.playground.create_debian_environments_collection()
        )
        self.task_data["binary_artifacts"] = [
            f"{binary_artifact.id}@collections"
        ]
        with self.assertRaisesRegex(
            LookupError,
            f"^'{binary_artifact.id}@collections' is of type 'collection' "
            f"instead of expected 'artifact_or_promise'",
        ):
            self.task

    def test_compute_dynamic_data_binary_promises(self) -> None:
        source_artifact = self.source_artifact

        with preserve_db_task_registry():

            class ExamplePipeline(
                SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
            ):
                """
                Pipeline workflow that provides a promise (e.g. by Sbuild).

                Piuparts will use the promise.
                """

                @override
                def populate(self) -> None:
                    sbuild = self.work_request.create_child_workflow(
                        task_name="sbuild",
                        task_data=SbuildWorkflowData(
                            input=SbuildInput(
                                source_artifact=source_artifact.id
                            ),
                            target_distribution="debian:sid",
                            architectures=["amd64"],
                        ),
                    )
                    sbuild = sbuild.create_child_worker(
                        task_name="sbuild",
                        task_data=SbuildData(
                            input=SbuildInput(
                                source_artifact=source_artifact.id
                            ),
                            build_architecture="amd64",
                            environment="debian/match:codename=sid",
                        ),
                    )
                    self.provides_artifact(
                        sbuild,
                        ArtifactCategory.BINARY_PACKAGE,
                        "build-arm64",
                        data={"source_package_name": "linux-base"},
                    )
                    self.provides_artifact(
                        sbuild,
                        ArtifactCategory.BINARY_PACKAGE,
                        "build-amd64",
                        data={
                            "architecture": "amd64",
                            "source_package_name": "hello",
                        },
                    )
                    self.provides_artifact(
                        sbuild,
                        ArtifactCategory.BINARY_PACKAGE,
                        "build-i386",
                        data={
                            "architecture": "i386",
                        },
                    )

                    self.work_request.create_child_workflow(
                        task_name="piuparts",
                        task_data=PiupartsWorkflowData(
                            source_artifact=source_artifact.id,
                            binary_artifacts=LookupMultiple(
                                (
                                    "internal@collections/name:build-amd64",
                                    "internal@collections/name:build-arm64",
                                    "internal@collections/name:build-i386",
                                )
                            ),
                            vendor="debian",
                            codename="sid",
                        ),
                    )

            root = self.playground.create_workflow(task_name="examplepipeline")
            self.assertTrue(orchestrate_workflow(root))

            piuparts = WorkRequest.objects.get(
                task_type=TaskTypes.WORKFLOW,
                task_name="piuparts",
            )

            computed_data = self.get_workflow(piuparts).compute_dynamic_data()
            self.assertEqual(
                computed_data,
                PiupartsWorkflowDynamicData(
                    subject="hello linux-base",
                    parameter_summary="hello linux-base in debian:sid",
                    source_artifact_id=source_artifact.id,
                    binary_artifacts_ids=[],
                ),
            )

    def test_get_input_artifacts(self) -> None:
        self.assertEqual(
            self.task.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=self.task.data.source_artifact,
                    label="source_artifact",
                    artifact_id=None,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=self.task.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=None,
                ),
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""View for QA work request."""

from functools import cached_property
from typing import Any, override

from debusine.artifacts.models import TaskTypes
from debusine.db.models import Artifact
from debusine.db.tasks import DBTask
from debusine.server.workflows import QAWorkflow
from debusine.tasks.models import RegressionAnalysisStatus
from debusine.web.views.work_request import Tab, WorkRequestPlugin


class QATab(Tab):
    """QA specialized tab."""

    def __init__(self, task: QAWorkflow) -> None:
        """Define this tab."""
        super().__init__(name="qa", label="QA")
        self.task = task
        self.work_request = task.work_request

    @cached_property
    def has_data(self) -> bool:
        """Check if the tab has data to display."""
        return (
            self.work_request.output_data is not None
            and self.work_request.output_data.regression_analysis is not None
            and "" in self.work_request.output_data.regression_analysis
        )

    @override
    def get_context_data(self) -> dict[str, Any]:
        # Assert the invariants that are checked by has_data
        assert self.work_request.output_data is not None
        regression_analysis = self.work_request.output_data.regression_analysis
        assert regression_analysis is not None
        summary = regression_analysis.get("")
        assert summary is not None

        # Prefetch referenced artifacts
        ids: set[int] = set()
        for analysis in regression_analysis.values():
            if analysis.original_artifact_id is not None:
                ids.add(analysis.original_artifact_id)
            if analysis.new_artifact_id is not None:
                ids.add(analysis.new_artifact_id)
        artifacts: dict[int, Artifact] = {}
        for artifact in Artifact.objects.filter(id__in=ids).select_related(
            "created_by_work_request"
        ):
            artifacts[artifact.pk] = artifact

        def get_artifact(pk: int | None) -> Artifact | None:
            if pk is None:
                return None
            return artifacts.get(pk)

        # TODO: is there a better way to get the package name?
        assert self.task.dynamic_data is not None
        assert self.task.dynamic_data.subject is not None
        package_name: str = self.task.dynamic_data.subject

        rows: list[dict[str, Any]] = []
        warnings: list[str] = []

        task = self.task
        assert isinstance(task, QAWorkflow)

        for name, analysis in sorted(regression_analysis.items()):
            if name == "":
                continue
            task_name, test_package_name, _ = name.split(":")
            if (
                task_name == "autopkgtest"
                and test_package_name != package_name
                and analysis.status
                in (
                    RegressionAnalysisStatus.NO_RESULT,
                    RegressionAnalysisStatus.STABLE,
                )
            ):
                # Hide NO_RESULT and STABLE reverse dependency autopkgtest
                # results to avoid cluttering the UI
                continue

            rows.append(
                {
                    "name": name,
                    "orig": get_artifact(analysis.original_artifact_id),
                    "new": get_artifact(analysis.new_artifact_id),
                    "analysis": analysis,
                }
            )

        return {
            "summary": summary,
            "orig": get_artifact(summary.original_artifact_id),
            "new": get_artifact(summary.new_artifact_id),
            "rows": rows,
            "warnings": warnings,
        }


class QAViewWorkRequestPlugin(WorkRequestPlugin):
    """View for QA work request."""

    task_type = TaskTypes.WORKFLOW
    task_name = "qa"
    task: QAWorkflow

    def __init__(self, task: DBTask[Any]) -> None:
        """Instantiate the tab."""
        super().__init__(task)
        self.tab = QATab(self.task)

    @override
    def is_enabled(self) -> bool:
        return self.tab.has_data

    @override
    def add_tabs(self, tabs: list[Tab]) -> None:
        tabs.insert(0, self.tab)

    @override
    def get_context_data(self) -> dict[str, Any]:
        return {}



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Template tag override_query_parameters."""

from typing import Any

from django import template

register = template.Library()


@register.simple_tag(takes_context=True)
def override_query_parameters(
    context: template.RequestContext, **kwargs: Any
) -> str:
    """
    Return query with overridden query parameters.

    If a query parameter is:
    page=1&order=id&asc=1

    Using this template tag with:
    override_query_parameters(page=2)

    result:
    page=2&order=id&asc=1
    """
    query = context.request.GET.copy()

    for k, v in kwargs.items():
        query[k] = v

    return query.urlencode()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command edit_worker_metadata."""

from typing import ClassVar, override

from debusine.db.models import Token, Worker
from debusine.django.management.tests import call_command
from debusine.test.django import TestCase


class EditWorkerMetadataCommandTests(TestCase):
    """Test for edit_worker_metadata command."""

    token: ClassVar[Token]
    worker: ClassVar[Worker]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Create a default Token and Worker."""
        super().setUpTestData()
        cls.token = cls.playground.create_worker_token(set_worker=False)
        cls.worker = Worker.objects.create_with_fqdn(
            "worker-01.lan", token=cls.token
        )

    def test_forwards_with_deprecation_warning(self) -> None:
        """Forwards to `worker edit_metadata` and warns."""
        static_metadata_file = self.create_temporary_file()
        static_metadata_file.write_bytes(
            b'sbuild:\n  architectures:\n    - amd64\n    - i386\n'
            b'  distributions:\n    - stretch\n    - buster'
        )
        with self.assertWarnsMessage(
            DeprecationWarning,
            "The `debusine-admin edit_worker_metadata` command has been "
            "deprecated in favour of `debusine-admin worker edit_metadata`",
        ):
            call_command(
                "edit_worker_metadata",
                "--set",
                str(static_metadata_file),
                self.worker.name,
            )

        self.worker.refresh_from_db()

        self.assertEqual(
            self.worker.static_metadata,
            {
                'sbuild': {
                    'architectures': ['amd64', 'i386'],
                    'distributions': ['stretch', 'buster'],
                }
            },
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views related to worker pools."""

from typing import Any, override

from django.db.models import Count, QuerySet
from django.shortcuts import get_object_or_404

from debusine.db.models import WorkerPool
from debusine.web.views.base import BaseUIView, DetailViewBase, ListViewBase
from debusine.web.views.places import Place
from debusine.web.views.table import TableMixin
from debusine.web.views.tables import WorkerPoolTable, WorkerTable


class WorkerPoolsListView(
    TableMixin[WorkerPool], BaseUIView, ListViewBase[WorkerPool]
):
    """List worker pools."""

    model = WorkerPool
    table_class = WorkerPoolTable
    base_template = "web/_base_debusine.html"
    template_name = "web/worker_pool-list.html"
    title = "List of worker pools"
    paginate_by = 20

    @override
    def get_queryset(self) -> QuerySet[WorkerPool, WorkerPool]:
        return super().get_queryset().annotate(workers=Count("worker"))


class WorkerPoolDetailView(BaseUIView, DetailViewBase[WorkerPool]):
    """Show details about a worker pool."""

    model = WorkerPool
    base_template = "web/_base_debusine.html"
    template_name = "web/worker_pool-detail.html"
    context_object_name = "worker_pool"

    @override
    def get_object(
        self, queryset: QuerySet[WorkerPool, WorkerPool] | None = None
    ) -> WorkerPool:
        """Return the Worker object to show."""
        assert queryset is None
        queryset = self.get_queryset()
        return get_object_or_404(queryset, name=self.kwargs["name"])

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)
        ctx["workers"] = WorkerTable(
            self.request,
            self.object.worker_set.all(),
        ).get_paginator(per_page=30)
        return ctx



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Infrastructure to create test scenarios in the database."""

import contextlib
import copy
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Any, Self, override

from debusine.db import playground
from debusine.db.models import File, FileStore
from debusine.server.file_backend.memory import MemoryFileBackend
from debusine.server.file_backend.models import (
    LocalFileBackendConfiguration,
    MemoryFileBackendConfiguration,
)


class Playground(playground.Playground, contextlib.ExitStack):
    """
    Generate test data scenarios for the database.

    It can be used with setUpTestData, in which case it supports the deepcopy
    protocol to preserve non-database assets across test methods
    """

    def __init__(self, memory_file_store: bool = True, **kwargs: Any) -> None:
        """Set default values."""
        super().__init__(**kwargs)
        self.file_store = self.get_default_file_store()
        self.memory_file_store = memory_file_store

    @override
    def __enter__(self) -> Self:
        """Set up the playground's file store."""
        super().__enter__()
        self.original_backend = self.file_store.backend
        self.original_configuration = self.file_store.configuration
        if self.memory_file_store:
            self.initial_storages: dict[str, dict[str, bytes]] | None = None
            self.file_store.backend = FileStore.BackendChoices.MEMORY
            mem_config = MemoryFileBackendConfiguration(
                name=self.file_store.name
            )
            self.file_store.configuration = mem_config.model_dump()
            self.file_store.save()
        else:
            self.stores_path = Path(
                self.enter_context(tempfile.TemporaryDirectory())
            )
            self.current_store_path = self.stores_path / "current"
            self.current_store_path.mkdir()
            self.saved_store_path = self.stores_path / "saved"
            local_config = LocalFileBackendConfiguration(
                base_directory=self.current_store_path.as_posix()
            )
            self.file_store.backend = FileStore.BackendChoices.LOCAL
            self.file_store.configuration = local_config.model_dump()
            self.file_store.save()
        return self

    @override
    def __exit__(self, *exc_details: Any) -> None:
        """Tear down the playground's file store."""
        self.file_store.backend = self.original_backend
        self.file_store.configuration = self.original_configuration
        self.file_store.save()
        for name in (
            "current_store_path",
            "initial_storages",
            "original_backend",
            "original_configuration",
            "saved_store_path",
            "stores_path",
        ):
            if hasattr(self, name):
                delattr(self, name)
        super().__exit__(*exc_details)

    def __deepcopy__(self, memo: dict[int, Any]) -> Self:
        """Deep copy of Playground, as used by Django's TestCase."""
        # Standard deepcopy
        cls = self.__class__
        result = cls.__new__(cls)
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))

        if hasattr(self, "original_backend"):
            match self.file_store.backend:
                case FileStore.BackendChoices.MEMORY:
                    self._deepcopy_memory(memo)
                case FileStore.BackendChoices.LOCAL:
                    self._deepcopy_files(memo)
                case _ as unreachable:
                    raise NotImplementedError(
                        f"backend {unreachable!r} not supported"
                    )

        return result

    def _deepcopy_memory(self, memo: dict[int, Any]) -> None:
        """Preserve the storage of MemoryFileBackend."""
        if self.initial_storages is None:
            self.initial_storages = copy.deepcopy(
                MemoryFileBackend.storages, memo
            )
        else:
            MemoryFileBackend.storages = copy.deepcopy(
                self.initial_storages, memo
            )

    def _deepcopy_files(
        self,
        memo: dict[int, Any],  # noqa: ARG002, U100
    ) -> None:
        """Preserve the storage of LocalFileBackend."""
        if not self.saved_store_path.exists():
            self.current_store_path.rename(self.saved_store_path)
        if self.current_store_path.exists():
            shutil.rmtree(self.current_store_path)
        # Hard link all files from saved_store to current_store
        subprocess.run(
            [
                "cp",
                "--link",
                "--recursive",
                self.saved_store_path.as_posix(),
                self.current_store_path,
            ],
            check=True,
        )

    @override
    def create_file_in_backend(self, *args: Any, **kwargs: Any) -> File:
        """
        Create a temporary file and adds it in the backend.

        :param backend: file backend to add the file in
        :param contents: contents of the file
        """
        file = super().create_file_in_backend(*args, **kwargs)
        if self.file_store.backend == FileStore.BackendChoices.LOCAL:
            backend = self.file_store.get_backend_object()
            self.callback(backend.remove_file, file)
        return file



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for MmDebstrap class."""

import copy
import shlex
from pathlib import Path
from typing import cast
from unittest import mock
from unittest.mock import call

import debusine.worker.tags as wtags
from debusine.artifacts.local_artifact import DebianSystemTarballArtifact
from debusine.artifacts.models import WorkRequestResults
from debusine.tasks import MmDebstrap
from debusine.tasks.models import (
    BaseDynamicTaskData,
    MmDebstrapBootstrapOptions,
    MmDebstrapData,
    WorkerType,
)
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT


class MmDebstrapTests(ExternalTaskHelperMixin[MmDebstrap], TestCase):
    """Unit tests for MmDebstrap class."""

    SAMPLE_TASK_DATA = {
        "bootstrap_options": {
            "architecture": "amd64",
            "extra_packages": ["hello", "python3"],
            "use_signed_by": True,
        },
        "bootstrap_repositories": [
            {
                "mirror": "https://deb.debian.org/deb",
                "suite": "bookworm",
                "components": ["main", "contrib"],
                "check_signature_with": "system",
            },
            {
                "types": ["deb-src"],
                "mirror": "https://example.com",
                "suite": "bullseye",
                "components": ["main"],
                "check_signature_with": "system",
                "keyring": {"url": "https://example.com/keyring.gpg"},
            },
        ],
    }

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            require_system={
                wtags.BUILD_ARCH_PREFIX + "amd64",
                wtags.CAP_PREFIX + "mmdebstrap",
            }
        )

    def test_compute_dynamic_data(self) -> None:
        """Test compute_dynamic_data."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            BaseDynamicTaskData(
                subject="bookworm",
                parameter_summary="bookworm (amd64)",
                runtime_context="amd64:None:hello,python3",
                configuration_context="amd64",
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "bookworm")

    def test_get_configuration_context(self) -> None:
        self.assertEqual(self.get_configuration_context(), "amd64")

    def test_configure_task(self) -> None:
        """self.configure_task() does not raise any exception."""
        self.configure_task(self.SAMPLE_TASK_DATA)

    def test_configure_task_defaults(self) -> None:
        """self.configure_task() set default values."""
        bootstrap_options = {
            "architecture": "amd64",
            "variant": "essential",
            "extra_packages": ["hello", "bye"],
        }

        self.configure_task(override={"bootstrap_options": bootstrap_options})

        self.assertTrue(self.task.data.bootstrap_options.use_signed_by)

    def test_get_provided_worker_tags(self) -> None:
        self.mock_is_command_available({"mmdebstrap": True})
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:mmdebstrap:version:1",
                wtags.CAP_PREFIX + "mmdebstrap",
            ],
        )

    def test_get_provided_worker_tags_mmdebstrap_not_available(self) -> None:
        self.mock_is_command_available({"mmdebstrap": False})
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [wtags.TASK_PREFIX + "worker:mmdebstrap:version:1"],
        )

    def test_analyze_worker(self) -> None:
        """Test the analyze_worker() method."""
        self.mock_is_command_available({"mmdebstrap": True})
        metadata = self.task.analyze_worker()
        self.assertEqual(metadata["mmdebstrap:available"], True)

    def test_analyze_worker_mmdebstrap_not_available(self) -> None:
        """analyze_worker() handles mmdebstrap not being available."""
        self.mock_is_command_available({"mmdebstrap": False})
        metadata = self.task.analyze_worker()
        self.assertEqual(metadata["mmdebstrap:available"], False)

    def test_can_run_on(self) -> None:
        """can_run_on returns True if mmdebstrap is available."""
        self.assertTrue(
            self.can_run_on(
                {
                    "system:architectures": ["amd64"],
                    "system:worker_type": WorkerType.EXTERNAL,
                    "mmdebstrap:available": True,
                    "mmdebstrap:version": self.task.TASK_VERSION,
                }
            )
        )

    def test_can_run_on_mismatched_task_version(self) -> None:
        """can_run_on returns False for mismatched task versions."""
        self.assertFalse(
            self.can_run_on(
                {
                    "system:architectures": ["amd64"],
                    "system:worker_type": WorkerType.EXTERNAL,
                    "mmdebstrap:available": True,
                    "mmdebstrap:version": self.task.TASK_VERSION + 1,
                }
            )
        )

    def test_can_run_on_missing_tool(self) -> None:
        """can_run_on returns False if mmdebstrap is not available."""
        self.assertFalse(
            self.can_run_on(
                {
                    "system:architectures": ["amd64"],
                    "system:worker_type": WorkerType.EXTERNAL,
                    "mmdebstrap:available": False,
                    "mmdebstrap:version": self.task.TASK_VERSION,
                }
            )
        )

    def test_can_run_on_wrong_architecture(self) -> None:
        """can_run_on returns False if the task is for a different arch."""
        self.assertFalse(
            self.can_run_on(
                {
                    "system:architectures": ["arm64"],
                    "system:worker_type": WorkerType.EXTERNAL,
                    "mmdebstrap:available": True,
                    "mmdebstrap:version": self.task.TASK_VERSION,
                }
            )
        )

    def test_cmdline_add_keyrings(self) -> None:
        """Command line has the keyring files."""
        self.task._host_sources_file = Path("/somewhere/some-file.sources")
        keyring_1 = self.create_temporary_file()
        keyring_2 = self.create_temporary_file()
        self.task._upload_keyrings = [keyring_1, keyring_2]
        self.task._keyrings = [keyring_1]

        cmdline = self.task._cmdline()
        self.assertIn(
            f"--customize-hook=upload {keyring_1} "
            f"/etc/apt/keyrings-debusine/{keyring_1.name}",
            cmdline,
        )
        self.assertIn(
            f"--customize-hook=upload {keyring_2} "
            f"/etc/apt/keyrings-debusine/{keyring_2.name}",
            cmdline,
        )

        self.assertIn(
            f"--keyring={keyring_1}",
            cmdline,
        )

        self.assertNotIn(
            f"--keyring={keyring_2}",
            cmdline,
        )

    def test_cmdline_minimum_options(self) -> None:
        """Command line has minimum options."""
        self.task._host_sources_file = Path("/somewhere/some-file.sources")
        self.task._chroot_sources_file = self.create_temporary_file()

        os_release_file = shlex.quote(self.task._OS_RELEASE_FILE)
        var_lib_dpkg = shlex.quote(self.task._VAR_LIB_DPKG)
        test_sbin_init = shlex.quote(self.task._TEST_SBIN_INIT_RETURN_CODE_FILE)

        expected = [
            "mmdebstrap",
            "--aptopt=Acquire::http::timeout=30",
            "--aptopt=Acquire::https::timeout=30",
            "--mode=unshare",
            "--format=tar",
            "--architectures=amd64",
            "--verbose",
            (
                '--customize-hook=cd "$1" && '
                "find etc/apt/sources.list.d -type f -delete"
            ),
            (
                f"--customize-hook=upload {self.task._chroot_sources_file} "
                "/etc/apt/sources.list.d/file.sources"
            ),
            '--customize-hook=mkdir "$1/etc/apt/keyrings-debusine"',
            f"--customize-hook=download /etc/os-release {os_release_file}",
            f"--customize-hook=copy-out /var/lib/dpkg {var_lib_dpkg}",
            '--customize-hook=rm -f "$1/etc/hostname"',
            (
                '--customize-hook='
                'test -h "$1/etc/resolv.conf" || rm -f "$1/etc/resolv.conf"'
            ),
            '--skip=chroot/start-stop-daemon',
            (
                '--customize-hook='
                '(test -x "$1/sbin/init" || test -h "$1/sbin/init") ; '
                'echo $? > "$1/test_sbin_init"'
            ),
            f'--customize-hook=download test_sbin_init {test_sbin_init}',
            '--customize-hook=rm "$1/test_sbin_init"',
            "--include=hello,python3",
            "",
            "system.tar.xz",
            "",
            str(self.task._host_sources_file),
        ]

        cmdline = self.task._cmdline()
        self.assertEqual(cmdline, expected)

    def test_cmdline_apt_timeout_tuple(self) -> None:
        """APT timeout uses read timeout when tuple is provided."""
        self.task._host_sources_file = Path("/somewhere/some-file.sources")
        self.task._chroot_sources_file = self.create_temporary_file()

        self.task.http_timeout = (5, 40)

        cmdline = self.task._cmdline()

        self.assertIn("--aptopt=Acquire::http::timeout=40", cmdline)
        self.assertIn("--aptopt=Acquire::https::timeout=40", cmdline)

    def test_cmdline_apt_timeout_int(self) -> None:
        """APT timeout uses integer value directly."""
        self.task._host_sources_file = Path("/somewhere/some-file.sources")
        self.task._chroot_sources_file = self.create_temporary_file()

        self.task.http_timeout = 20

        cmdline = self.task._cmdline()

        self.assertIn("--aptopt=Acquire::http::timeout=20", cmdline)
        self.assertIn("--aptopt=Acquire::https::timeout=20", cmdline)

    def test_cmdline_apt_timeout_default(self) -> None:
        """APT timeout falls back to DEFAULT_REQUEST_TIMEOUT."""
        self.task._host_sources_file = Path("/somewhere/some-file.sources")
        self.task._chroot_sources_file = self.create_temporary_file()

        self.task.http_timeout = None

        cmdline = self.task._cmdline()

        expected = DEFAULT_REQUEST_TIMEOUT[1]

        self.assertIn(f"--aptopt=Acquire::http::timeout={expected}", cmdline)
        self.assertIn(f"--aptopt=Acquire::https::timeout={expected}", cmdline)

    def test_cmdline_variant(self) -> None:
        """_cmdline() include --variant=bootstrap_options['variant']."""
        bootstrap_options = {
            "architecture": "amd64",
            "variant": "essential",
        }

        self.configure_task(override={"bootstrap_options": bootstrap_options})
        self.task._host_sources_file = Path("some-file.sources")
        cmdline = self.task._cmdline()
        self.assertIn("--variant=essential", cmdline)

    def test_cmdline_no_extra_packages(self) -> None:
        """_cmdline() without extra_packages."""
        bootstrap_options = {
            "architecture": "amd64",
            "variant": "essential",
            "use_signed_by": True,
        }

        self.configure_task(override={"bootstrap_options": bootstrap_options})
        self.task._host_sources_file = Path("some-file.sources")
        cmdline = self.task._cmdline()
        for arg in cmdline:
            self.assertFalse(arg.startswith("--include="))

    def test_cmdline_extra_packages(self) -> None:
        """_cmdline() include --include=extra_packages."""
        bootstrap_options = {
            "architecture": "amd64",
            "variant": "essential",
            "extra_packages": ["hello", "bye"],
            "use_signed_by": True,
        }

        self.configure_task(override={"bootstrap_options": bootstrap_options})
        self.task._host_sources_file = Path("some-file.sources")
        cmdline = self.task._cmdline()
        self.assertIn("--include=hello,bye", cmdline)

    def test_cmdline_gpg_command(self) -> None:
        """_cmdline() include APT configuration for gpg_command."""
        bootstrap_options = {
            "architecture": "amd64",
            "gpgv_command": "gpgv",
        }

        self.configure_task(override={"bootstrap_options": bootstrap_options})
        self.task._host_sources_file = Path("some-file.sources")
        cmdline = self.task._cmdline()
        self.assertIn('--aptopt=Apt::Key::gpgvcommand "/usr/bin/gpgv"', cmdline)

    def test_cmdline_keyring_package(self) -> None:
        """_cmdline() include --include=keyring_0 --include=keyring_1."""
        task_data = copy.deepcopy(self.SAMPLE_TASK_DATA)

        bootstrap_repos = cast(
            list[dict[str, str]], task_data["bootstrap_repositories"]
        )
        bootstrap_repos[0]["keyring_package"] = "keyring_0"
        bootstrap_repos[1]["keyring_package"] = "keyring_1"

        self.configure_task(task_data)
        self.task._host_sources_file = Path("some-file.sources")
        cmdline = self.task._cmdline()

        self.assertIn("--include=keyring_0", cmdline)
        self.assertIn("--include=keyring_1", cmdline)

    def test_cmdline_customization_script(self) -> None:
        """_cmdline() include customization_script arguments."""
        self.task._customization_script = (
            self.create_temporary_directory() / "customization_script"
        )
        cmdline = self.task._cmdline()

        customization_script = self.task._customization_script
        script_name = "customization_script"
        self.assertIn(
            f"--customize-hook=upload {customization_script} /{script_name}",
            cmdline,
        )
        self.assertIn(f'--customize-hook=chmod 555 "$1/{script_name}"', cmdline)
        self.assertIn(f'--customize-hook=chroot "$1" /{script_name}', cmdline)
        self.assertIn(f'--customize-hook=rm "$1/{script_name}"', cmdline)

        self.assertIn("", cmdline)

    def test_fetch_input(self) -> None:
        """Test fetch_input method."""
        # Directory does not need to exist: it is not used
        directory = Path()
        self.assertTrue(self.task.fetch_input(directory))

    def test_upload_artifacts(self) -> None:
        """Test upload_artifacts()."""
        directory = self.create_temporary_directory()

        system_tarball = directory / MmDebstrap._OUTPUT_SYSTEM_FILE
        system_tarball.write_bytes(b"Generated tarball")

        # Debusine.upload_artifact is mocked to verify the call only
        debusine_mock = self.mock_debusine()

        os_release_data = self.write_os_release(
            directory / self.task._OS_RELEASE_FILE
        )
        (directory / self.task._TEST_SBIN_INIT_RETURN_CODE_FILE).write_text("0")

        packages = self.patch_subprocess_run_pkglist()

        mirror = (
            MmDebstrapData.model_validate(self.SAMPLE_TASK_DATA)
            .bootstrap_repositories[0]
            .mirror
        )
        bootstrap_options = MmDebstrapBootstrapOptions.model_validate(
            self.SAMPLE_TASK_DATA["bootstrap_options"]
        )

        self.task.upload_artifacts(
            directory, execution_result=WorkRequestResults.SUCCESS
        )

        calls = []

        expected_system_artifact = DebianSystemTarballArtifact.create(
            system_tarball,
            data={
                "variant": None,
                "architecture": bootstrap_options.architecture,
                "vendor": os_release_data["ID"],
                "codename": os_release_data["VERSION_CODENAME"],
                "components": ["main", "contrib"],
                "pkglist": packages,
                "with_dev": True,
                "with_init": True,
                "mirror": mirror,
            },
        )

        calls.append(
            call(
                expected_system_artifact,
                workspace=self.task.workspace_name,
                work_request=self.task.work_request_id,
            )
        )

        debusine_mock.upload_artifact.assert_has_calls(calls)

    def test_upload_artifacts_sid(self) -> None:
        """Test upload_artifacts() for a sid tarball."""
        directory = self.create_temporary_directory()

        system_tarball = directory / MmDebstrap._OUTPUT_SYSTEM_FILE
        system_tarball.write_bytes(b"Generated tarball")

        # Debusine.upload_artifact is mocked to verify the call only
        debusine_mock = self.mock_debusine()

        os_release_data = self.write_os_release(
            directory / self.task._OS_RELEASE_FILE,
            codename="trixie",
            version=13,
        )
        (directory / self.task._TEST_SBIN_INIT_RETURN_CODE_FILE).write_text("0")

        packages = self.patch_subprocess_run_pkglist()

        mirror = (
            MmDebstrapData.model_validate(self.SAMPLE_TASK_DATA)
            .bootstrap_repositories[0]
            .mirror
        )
        bootstrap_options = MmDebstrapBootstrapOptions.model_validate(
            self.SAMPLE_TASK_DATA["bootstrap_options"]
        )
        self.task.data.bootstrap_repositories[0].suite = "unstable"

        self.task.upload_artifacts(
            directory, execution_result=WorkRequestResults.SUCCESS
        )

        calls = []

        expected_system_artifact = DebianSystemTarballArtifact.create(
            system_tarball,
            data={
                "variant": None,
                "architecture": bootstrap_options.architecture,
                "vendor": os_release_data["ID"],
                "codename": "sid",
                "components": ["main", "contrib"],
                "pkglist": packages,
                "with_dev": True,
                "with_init": True,
                "mirror": mirror,
            },
        )

        calls.append(
            call(
                expected_system_artifact,
                workspace=self.task.workspace_name,
                work_request=self.task.work_request_id,
            )
        )

        debusine_mock.upload_artifact.assert_has_calls(calls)

    def test_upload_artifacts_do_nothing(self) -> None:
        """Test upload_artifacts() doing nothing: execution_result=FAILURE."""
        self.mock_debusine()
        self.task.upload_artifacts(
            Path(), execution_result=WorkRequestResults.FAILURE
        )

    def test_upload_no_components_specified(self) -> None:
        """Test upload_artifacts() falls back to sources_file's components."""
        directory = self.create_temporary_directory()

        mirror = "http://deb.debian.org/debian"
        self.configure_task(
            override={
                "bootstrap_repositories": [
                    {"mirror": mirror, "suite": "unstable"}
                ]
            }
        )
        components = ["main", "contrib"]
        with mock.patch(
            "debusine.tasks.systembootstrap.SystemBootstrap."
            "_list_components_for_suite",
            return_value=components,
            autospec=True,
        ):
            self.task.configure_for_execution(directory)

        system_tarball = directory / MmDebstrap._OUTPUT_SYSTEM_FILE
        system_tarball.write_bytes(b"Generated tarball")

        # Debusine.upload_artifact is mocked to verify the call only
        debusine_mock = self.mock_debusine()

        os_release_data = self.write_os_release(
            directory / self.task._OS_RELEASE_FILE
        )
        (directory / self.task._TEST_SBIN_INIT_RETURN_CODE_FILE).write_text("0")

        packages = self.patch_subprocess_run_pkglist()

        self.task.upload_artifacts(
            directory, execution_result=WorkRequestResults.SUCCESS
        )

        calls = []

        bootstrap_options = MmDebstrapBootstrapOptions.model_validate(
            self.SAMPLE_TASK_DATA["bootstrap_options"]
        )
        expected_system_artifact = DebianSystemTarballArtifact.create(
            system_tarball,
            data={
                "variant": None,
                "architecture": bootstrap_options.architecture,
                "vendor": os_release_data["ID"],
                "codename": "sid",
                "components": components,
                "pkglist": packages,
                "with_dev": True,
                "with_init": True,
                "mirror": mirror,
            },
        )

        calls.append(
            call(
                expected_system_artifact,
                workspace=self.task.workspace_name,
                work_request=self.task.work_request_id,
            )
        )

        debusine_mock.upload_artifact.assert_has_calls(calls)

    def test_variant_enums(self) -> None:
        """Check MmDebstrap variants enums."""
        systembootstrap_variants = ["buildd", "minbase"]
        mmdebstrap_variants = [
            "-",
            "apt",
            "custom",
            "debootstrap",
            "essential",
            "extract",
            "important",
            "required",
            "standard",
        ]

        for variant in systembootstrap_variants + mmdebstrap_variants:
            with self.subTest(variant=variant):
                self.configure_task(
                    override={
                        "bootstrap_options": {
                            "architecture": "amd64",
                            "variant": variant,
                            "extra_packages": ["hello"],
                            "use_signed_by": True,
                        }
                    }
                )

    def patch_subprocess_run_pkglist(self) -> dict[str, str]:
        """Patch subprocess.run() and return dictionary with name -> version."""
        mock_result = mock.MagicMock()
        packages = {"aardvark-dns:amd64": "1.0.4", "abcde": "1:2.4.3"}

        mock_result.stdout = (
            "\n".join([f"{pkg}\t{ver}" for pkg, ver in packages.items()]) + "\n"
        )

        patch = mock.patch(
            "subprocess.run", autospec=True, return_value=mock_result
        )
        patch.start()
        self.addCleanup(patch.stop)

        return packages

    def test_get_with_init(self) -> None:
        """Test _get_with_init()."""
        file = self.create_temporary_file(contents=b"0\n")
        self.assertTrue(self.task._get_with_init(file))

        file.write_text("1\n")
        self.assertFalse(self.task._get_with_init(file))



# Generated by Django 5.2.13 on 2026-05-05 16:39

from django.db import migrations, models

import debusine.db.models.auth


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0055_system_apt_credentials'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='token',
            name='db_token_valid_type',
        ),
        migrations.AddConstraint(
            model_name='token',
            constraint=models.CheckConstraint(
                check=models.Q(
                    (
                        'token_type__in',
                        [
                            debusine.db.models.auth._TokenTypes.USER,
                            debusine.db.models.auth._TokenTypes.WORK_REQUEST,
                            debusine.db.models.auth._TokenTypes.WORKER,
                            debusine.db.models.auth._TokenTypes.WORKER_ACTIVATION,
                        ],
                    )
                ),
                name='db_token_valid_type',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

import copy
from collections.abc import Collection, Sequence
from functools import partial
from typing import Any

from django.db import migrations, transaction
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps
from django.db.models import F, Q

from debusine.artifacts.models import TaskTypes


def _dict_has_path(d: dict[str, Any], path: Sequence[str]) -> bool:
    """Return whether `d[path[0]][path[1]][...]` exists."""
    for key in path[:-1]:  # pragma: no cover
        d = d.get(key, {})
    return path[-1] in d


def _dict_pop_path(d: dict[str, Any], path: Sequence[str]) -> Any:
    """Pop and return `d[path[0]][path[1]][...]`."""
    for key in path[:-1]:  # pragma: no cover
        d = d[key]
    return d.pop(path[-1])


def _dict_set_path(d: dict[str, Any], path: Sequence[str], value: Any) -> None:
    """Set `d[path[0]][path[1]][...]` to `value`."""
    for key in path[:-1]:  # pragma: no cover
        d = d.setdefault(key, {})
    d[path[-1]] = value


def make_data_field_renamer(
    model_name: str,
    data_field: str,
    filters: dict[str, Any],
    renames: list[tuple[str, str]],
) -> migrations.RunPython:
    """
    Make a migration operation to rename keys in data of model_name.

    This will be slow if there are many matching models.  Let's hope there
    aren't yet.
    """

    def operation(
        model_name: str,
        data_field: str,
        filters: dict[str, Any],
        renames: list[tuple[str, str]],
        apps: StateApps,
        schema_editor: BaseDatabaseSchemaEditor,  # noqa: ARG001, U100
    ) -> None:
        Model = apps.get_model("db", model_name)
        rename_paths = [
            (old_name.split("__"), new_name.split("__"))
            for old_name, new_name in renames
        ]
        path_filter = Q()
        for old_name, _ in renames:
            path_filter |= Q(
                **{f"{data_field}__has_key": F(f"{data_field}__{old_name}")}
            )
        while True:
            with transaction.atomic():
                models = Model.objects.select_for_update().filter(
                    path_filter, **filters
                )[:1000]
                if not models:
                    break
                for model in models:
                    data = copy.deepcopy(getattr(model, data_field))
                    for old_path, new_path in rename_paths:
                        try:
                            old_value = _dict_pop_path(data, old_path)
                        except KeyError:  # pragma: no cover
                            pass
                        else:
                            assert not _dict_has_path(data, new_path)
                            _dict_set_path(data, new_path, old_value)
                    setattr(model, data_field, data)
                    model.save()

    return migrations.RunPython(
        partial(operation, model_name, data_field, filters, renames),
        reverse_code=partial(
            operation,
            model_name,
            data_field,
            filters,
            [(new_name, old_name) for old_name, new_name in renames],
        ),
    )


def make_work_request_task_data_field_renamer(
    task_type: TaskTypes,
    task_name: str | Collection[str],
    renames: list[tuple[str, str]],
) -> migrations.RunPython:
    """Make a migration operation to rename task_data in a WorkRequest."""
    filters: dict[str, Any] = {"task_type": task_type}
    if isinstance(task_name, str):
        filters["task_name"] = task_name
    else:
        filters["task_name__in"] = task_name
    return make_data_field_renamer("WorkRequest", "task_data", filters, renames)


def make_workflow_template_task_data_field_renamer(
    task_name: str | Collection[str], renames: list[tuple[str, str]]
) -> migrations.RunPython:
    """Make a migration operation to rename task_data in a WorkflowTemplate."""
    filters: dict[str, Any] = {}
    if isinstance(task_name, str):
        filters["task_name"] = task_name
    else:
        filters["task_name__in"] = task_name
    return make_data_field_renamer(
        "WorkflowTemplate", "task_data", filters, renames
    )
